import itertools
import json
import random
import ssl
from contextlib import contextmanager
from datetime import timedelta
from pickle import dumps, loads
from unittest.mock import ANY, Mock, call, patch

import pytest

try:
    from redis import CredentialProvider, exceptions
except ImportError:
    exceptions = None
    CredentialProvider = None

from celery import signature, states, uuid
from celery.canvas import Signature
from celery.contrib.testing.mocks import ContextMock
from celery.exceptions import BackendStoreError, ChordError, ImproperlyConfigured
from celery.result import AsyncResult, GroupResult
from celery.utils.collections import AttributeDict
from t.unit import conftest


def raise_on_second_call(mock, exc, *retval):
    def on_first_call(*args, **kwargs):
        mock.side_effect = exc
        return mock.return_value

    mock.side_effect = on_first_call
    if retval:
        mock.return_value, = retval


class ConnectionError(Exception):
    pass


class Connection:
    connected = True

    def disconnect(self):
        self.connected = False


class Pipeline:
    def __init__(self, client):
        self.client = client
        self.steps = []

    def __getattr__(self, attr):
        def add_step(*args, **kwargs):
            self.steps.append((getattr(self.client, attr), args, kwargs))
            return self

        return add_step

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        pass

    def execute(self):
        return [step(*a, **kw) for step, a, kw in self.steps]


class PubSub(conftest.MockCallbacks):
    def __init__(self, ignore_subscribe_messages=False):
        self._subscribed_to = set()

    def close(self):
        self._subscribed_to = set()

    def subscribe(self, *args):
        self._subscribed_to.update(args)

    def unsubscribe(self, *args):
        self._subscribed_to.difference_update(args)

    def get_message(self, timeout=None):
        pass


class Redis(conftest.MockCallbacks):
    Connection = Connection
    Pipeline = Pipeline
    pubsub = PubSub

    def __init__(self, host=None, port=None, db=None, password=None, **kw):
        self.host = host
        self.port = port
        self.db = db
        self.password = password
        self.keyspace = {}
        self.expiry = {}
        self.connection = self.Connection()

    def get(self, key):
        return self.keyspace.get(key)

    def mget(self, keys):
        return [self.get(key) for key in keys]

    def setex(self, key, expires, value):
        self.set(key, value)
        self.expire(key, expires)

    def set(self, key, value):
        self.keyspace[key] = value

    def expire(self, key, expires):
        self.expiry[key] = expires
        return expires

    def delete(self, key):
        return bool(self.keyspace.pop(key, None))

    def pipeline(self):
        return self.Pipeline(self)

    def _get_unsorted_list(self, key):
        # We simply store the values in append (rpush) order
        return self.keyspace.setdefault(key, list())

    def rpush(self, key, value):
        self._get_unsorted_list(key).append(value)

    def lrange(self, key, start, stop):
        return self._get_unsorted_list(key)[start:stop]

    def llen(self, key):
        return len(self._get_unsorted_list(key))

    def _get_sorted_set(self, key):
        # We store 2-tuples of (score, value) and sort after each append (zadd)
        return self.keyspace.setdefault(key, list())

    def zadd(self, key, mapping):
        # Store elements as 2-tuples with the score first so we can sort it
        # once the new items have been inserted
        fake_sorted_set = self._get_sorted_set(key)
        fake_sorted_set.extend(
            (score, value) for value, score in mapping.items()
        )
        fake_sorted_set.sort()

    def zrange(self, key, start, stop):
        # `stop` is inclusive in Redis so we use `stop + 1` unless that would
        # cause us to move from negative (right-most) indices to positive
        stop = stop + 1 if stop != -1 else None
        return [e[1] for e in self._get_sorted_set(key)[start:stop]]

    def zrangebyscore(self, key, min_, max_):
        return [
            e[1] for e in self._get_sorted_set(key)
            if (min_ == "-inf" or e[0] >= min_) and
            (max_ == "+inf" or e[1] <= max_)
        ]

    def zcount(self, key, min_, max_):
        return len(self.zrangebyscore(key, min_, max_))


class Sentinel(conftest.MockCallbacks):
    def __init__(self, sentinels, min_other_sentinels=0, sentinel_kwargs=None,
                 **connection_kwargs):
        self.sentinel_kwargs = sentinel_kwargs
        self.sentinels = [Redis(hostname, port, **self.sentinel_kwargs)
                          for hostname, port in sentinels]
        self.min_other_sentinels = min_other_sentinels
        self.connection_kwargs = connection_kwargs

    def master_for(self, service_name, redis_class, **kwargs):
        self.master_for_kwargs = kwargs
        return random.choice(self.sentinels)


class redis:
    StrictRedis = Redis

    class ConnectionPool:
        def __init__(self, **kwargs):
            pass

    class UnixDomainSocketConnection:
        def __init__(self, **kwargs):
            pass


class sentinel:
    Sentinel = Sentinel


class test_RedisResultConsumer:
    def get_backend(self):
        from celery.backends.redis import RedisBackend

        class _RedisBackend(RedisBackend):
            redis = redis

        return _RedisBackend(app=self.app)

    def get_consumer(self):
        consumer = self.get_backend().result_consumer
        consumer._connection_errors = (ConnectionError,)
        return consumer

    @patch('celery.backends.asynchronous.BaseResultConsumer.on_after_fork')
    def test_on_after_fork(self, parent_method):
        consumer = self.get_consumer()
        consumer.start('none')
        consumer.on_after_fork()
        parent_method.assert_called_once()
        consumer.backend.client.connection_pool.reset.assert_called_once()
        consumer._pubsub.close.assert_called_once()
        # PubSub instance not initialized - exception would be raised
        # when calling .close()
        consumer._pubsub = None
        parent_method.reset_mock()
        consumer.backend.client.connection_pool.reset.reset_mock()
        consumer.on_after_fork()
        parent_method.assert_called_once()
        consumer.backend.client.connection_pool.reset.assert_called_once()

        # Continues on KeyError
        consumer._pubsub = Mock()
        consumer._pubsub.close = Mock(side_effect=KeyError)
        parent_method.reset_mock()
        consumer.backend.client.connection_pool.reset.reset_mock()
        consumer.on_after_fork()
        parent_method.assert_called_once()

    @patch('celery.backends.redis.ResultConsumer.cancel_for')
    @patch('celery.backends.asynchronous.BaseResultConsumer.on_state_change')
    def test_on_state_change(self, parent_method, cancel_for):
        consumer = self.get_consumer()
        meta = {'task_id': 'testing', 'status': states.SUCCESS}
        message = 'hello'
        consumer.on_state_change(meta, message)
        parent_method.assert_called_once_with(meta, message)
        cancel_for.assert_called_once_with(meta['task_id'])

        # Does not call cancel_for for other states
        meta = {'task_id': 'testing2', 'status': states.PENDING}
        parent_method.reset_mock()
        cancel_for.reset_mock()
        consumer.on_state_change(meta, message)
        parent_method.assert_called_once_with(meta, message)
        cancel_for.assert_not_called()

    def test_drain_events_before_start(self):
        consumer = self.get_consumer()
        # drain_events shouldn't crash when called before start
        consumer.drain_events(0.001)

    def test_consume_from_connection_error(self):
        consumer = self.get_consumer()
        consumer.start('initial')
        consumer._pubsub.subscribe.side_effect = (ConnectionError(), None)
        consumer.consume_from('some-task')
        assert consumer._pubsub._subscribed_to == {b'celery-task-meta-initial', b'celery-task-meta-some-task'}

    def test_cancel_for_connection_error(self):
        consumer = self.get_consumer()
        consumer.start('initial')
        consumer._pubsub.unsubscribe.side_effect = ConnectionError()
        consumer.consume_from('some-task')
        consumer.cancel_for('some-task')
        assert consumer._pubsub._subscribed_to == {b'celery-task-meta-initial'}

    @patch('celery.backends.redis.ResultConsumer.cancel_for')
    @patch('celery.backends.asynchronous.BaseResultConsumer.on_state_change')
    def test_drain_events_connection_error(self, parent_on_state_change, cancel_for):
        meta = {'task_id': 'initial', 'status': states.SUCCESS}
        consumer = self.get_consumer()
        consumer.start('initial')
        consumer.backend._set_with_state(b'celery-task-meta-initial', json.dumps(meta), states.SUCCESS)
        consumer._pubsub.get_message.side_effect = ConnectionError()
        consumer.drain_events()
        parent_on_state_change.assert_called_with(meta, None)
        assert consumer._pubsub._subscribed_to == {b'celery-task-meta-initial'}

    def test_drain_events_connection_error_no_patch(self):
        meta = {'task_id': 'initial', 'status': states.SUCCESS}
        consumer = self.get_consumer()
        consumer.start('initial')
        consumer.backend._set_with_state(b'celery-task-meta-initial', json.dumps(meta), states.SUCCESS)
        consumer._pubsub.get_message.side_effect = ConnectionError()
        consumer.drain_events()
        consumer._pubsub.subscribe.assert_not_called()

    def test__reconnect_pubsub_no_subscribed(self):
        consumer = self.get_consumer()
        consumer.start('initial')
        consumer.subscribed_to = set()
        consumer._reconnect_pubsub()
        consumer.backend.client.mget.assert_not_called()
        consumer._pubsub.subscribe.assert_not_called()
        consumer._pubsub.connection.register_connect_callback.assert_called_once()

    def test__reconnect_pubsub_with_state_change(self):
        meta = {'task_id': 'initial', 'status': states.SUCCESS}
        consumer = self.get_consumer()
        consumer.start('initial')
        consumer.backend._set_with_state(b'celery-task-meta-initial', json.dumps(meta), states.SUCCESS)
        consumer._reconnect_pubsub()
        consumer.backend.client.mget.assert_called_once()
        consumer._pubsub.subscribe.assert_not_called()
        consumer._pubsub.connection.register_connect_callback.assert_called_once()

    def test__reconnect_pubsub_without_state_change(self):
        meta = {'task_id': 'initial', 'status': states.STARTED}
        consumer = self.get_consumer()
        consumer.start('initial')
        consumer.backend._set_with_state(b'celery-task-meta-initial', json.dumps(meta), states.SUCCESS)
        consumer._reconnect_pubsub()
        consumer.backend.client.mget.assert_called_once()
        consumer._pubsub.subscribe.assert_called_once()
        consumer._pubsub.connection.register_connect_callback.assert_not_called()


class basetest_RedisBackend:
    def get_backend(self):
        from celery.backends.redis import RedisBackend

        class _RedisBackend(RedisBackend):
            redis = redis

        return _RedisBackend

    def get_E_LOST(self):
        from celery.backends.redis import E_LOST
        return E_LOST

    def create_task(self, i, group_id="group_id"):
        tid = uuid()
        task = Mock(name=f'task-{tid}')
        task.name = 'foobarbaz'
        self.app.tasks['foobarbaz'] = task
        task.request.chord = signature(task)
        task.request.id = tid
        self.b.set_chord_size(group_id, 10)
        task.request.group = group_id
        task.request.group_index = i
        return task

    @contextmanager
    def chord_context(self, size=1):
        with patch('celery.backends.redis.maybe_signature') as ms:
            request = Mock(name='request')
            request.id = 'id1'
            group_id = 'gid1'
            request.group = group_id
            request.group_index = None
            tasks = [
                self.create_task(i, group_id=request.group)
                for i in range(size)
            ]
            callback = ms.return_value = Signature('add')
            callback.id = 'id1'
            self.b.set_chord_size(group_id, size)
            callback.delay = Mock(name='callback.delay')
            yield tasks, request, callback

    def setup_method(self):
        self.Backend = self.get_backend()
        self.E_LOST = self.get_E_LOST()
        self.b = self.Backend(app=self.app)


class MyCredentialProvider(CredentialProvider):
    pass


class NonCredentialProvider:
    pass


class test_RedisBackend(basetest_RedisBackend):
    @pytest.mark.usefixtures('depends_on_current_app')
    def test_reduce(self):
        pytest.importorskip('redis')

        from celery.backends.redis import RedisBackend
        x = RedisBackend(app=self.app)
        assert loads(dumps(x))

    def test_no_redis(self):
        self.Backend.redis = None
        with pytest.raises(ImproperlyConfigured):
            self.Backend(app=self.app)

    def test_username_password_from_redis_conf(self):
        self.app.conf.redis_password = 'password'
        x = self.Backend(app=self.app)

        assert x.connparams
        assert 'username' not in x.connparams
        assert x.connparams['password'] == 'password'
        self.app.conf.redis_username = 'username'
        x = self.Backend(app=self.app)

        assert x.connparams
        assert x.connparams['username'] == 'username'
        assert x.connparams['password'] == 'password'

    def test_credential_provider_from_redis_conf(self):
        self.app.conf.redis_backend_credential_provider = "redis.CredentialProvider"
        x = self.Backend(app=self.app)

        assert x.connparams
        assert 'credential_provider' in x.connparams
        assert 'username' not in x.connparams
        assert 'password' not in x.connparams

        # with local credential provider
        self.app.conf.redis_backend_credential_provider = MyCredentialProvider()
        x = self.Backend(app=self.app)
        assert x.connparams
        assert 'credential_provider' in x.connparams
        assert 'username' not in x.connparams
        assert 'password' not in x.connparams

        # raise ImportError
        self.app.conf.redis_backend_credential_provider = "not_exist.CredentialProvider"
        with pytest.raises(ImportError):
            self.Backend(app=self.app)

        # raise value Error
        self.app.conf.redis_backend_credential_provider = NonCredentialProvider()
        with pytest.raises(ValueError):
            self.Backend(app=self.app)

    def test_url(self):
        self.app.conf.redis_socket_timeout = 30.0
        self.app.conf.redis_socket_connect_timeout = 100.0
        x = self.Backend(
            'redis://:bosco@vandelay.com:123//1', app=self.app,
        )
        assert x.connparams
        assert x.connparams['host'] == 'vandelay.com'
        assert x.connparams['db'] == 1
        assert x.connparams['port'] == 123
        assert x.connparams['password'] == 'bosco'
        assert x.connparams['socket_timeout'] == 30.0
        assert x.connparams['socket_connect_timeout'] == 100.0
        assert 'username' not in x.connparams

        x = self.Backend(
            'redis://username:bosco@vandelay.com:123//1', app=self.app,
        )
        assert x.connparams
        assert x.connparams['host'] == 'vandelay.com'
        assert x.connparams['db'] == 1
        assert x.connparams['port'] == 123
        assert x.connparams['username'] == 'username'
        assert x.connparams['password'] == 'bosco'
        assert x.connparams['socket_timeout'] == 30.0
        assert x.connparams['socket_connect_timeout'] == 100.0

    def test_url_with_credential_provider(self):
        self.app.conf.redis_socket_timeout = 30.0
        self.app.conf.redis_socket_connect_timeout = 100.0
        x = self.Backend(
            'redis://:bosco@vandelay.com:123/1?credential_provider=redis.CredentialProvider', app=self.app,
        )

        assert x.connparams
        assert x.connparams['host'] == 'vandelay.com'
        assert x.connparams['db'] == 1
        assert x.connparams['port'] == 123
        assert x.connparams['socket_timeout'] == 30.0
        assert x.connparams['socket_connect_timeout'] == 100.0
        assert isinstance(x.connparams['credential_provider'], CredentialProvider)
        assert "username" not in x.connparams
        assert "password" not in x.connparams

        # without username and password
        x = self.Backend(
            'redis://@vandelay.com:123/1?credential_provider=redis.UsernamePasswordCredentialProvider', app=self.app,
        )
        assert x.connparams
        assert x.connparams['host'] == 'vandelay.com'
        assert x.connparams['db'] == 1
        assert x.connparams['port'] == 123
        assert isinstance(x.connparams['credential_provider'], CredentialProvider)

        # raise importError
        with pytest.raises(ImportError):
            self.Backend(
                'redis://@vandelay.com:123/1?credential_provider=not_exist.CredentialProvider', app=self.app,
            )

        # raise valueError
        with pytest.raises(ValueError):
            # some non-credential provider class
            # not ideal but serve purpose
            self.Backend(
                'redis://@vandelay.com:123/1?credential_provider=abc.ABC', app=self.app,
            )

    def test_timeouts_in_url_coerced(self):
        pytest.importorskip('redis')

        x = self.Backend(
            ('redis://:bosco@vandelay.com:123//1?'
             'socket_timeout=30&socket_connect_timeout=100'),
            app=self.app,
        )
        assert x.connparams
        assert x.connparams['host'] == 'vandelay.com'
        assert x.connparams['db'] == 1
        assert x.connparams['port'] == 123
        assert x.connparams['password'] == 'bosco'
        assert x.connparams['socket_timeout'] == 30
        assert x.connparams['socket_connect_timeout'] == 100

    def test_socket_url(self):
        pytest.importorskip('redis')

        self.app.conf.redis_socket_timeout = 30.0
        self.app.conf.redis_socket_connect_timeout = 100.0
        x = self.Backend(
            'socket:///tmp/redis.sock?virtual_host=/3', app=self.app,
        )
        assert x.connparams
        assert x.connparams['path'] == '/tmp/redis.sock'
        assert (x.connparams['connection_class'] is
                redis.UnixDomainSocketConnection)
        assert 'host' not in x.connparams
        assert 'port' not in x.connparams
        assert x.connparams['socket_timeout'] == 30.0
        assert 'socket_connect_timeout' not in x.connparams
        assert 'socket_keepalive' not in x.connparams
        assert x.connparams['db'] == 3

    def test_backend_ssl(self):
        pytest.importorskip('redis')

        self.app.conf.redis_backend_use_ssl = {
            'ssl_cert_reqs': ssl.CERT_REQUIRED,
            'ssl_ca_certs': '/path/to/ca.crt',
            'ssl_certfile': '/path/to/client.crt',
            'ssl_keyfile': '/path/to/client.key',
        }
        self.app.conf.redis_socket_timeout = 30.0
        self.app.conf.redis_socket_connect_timeout = 100.0
        x = self.Backend(
            'rediss://:bosco@vandelay.com:123//1', app=self.app,
        )
        assert x.connparams
        assert x.connparams['host'] == 'vandelay.com'
        assert x.connparams['db'] == 1
        assert x.connparams['port'] == 123
        assert x.connparams['password'] == 'bosco'
        assert x.connparams['socket_timeout'] == 30.0
        assert x.connparams['socket_connect_timeout'] == 100.0
        assert x.connparams['ssl_cert_reqs'] == ssl.CERT_REQUIRED
        assert x.connparams['ssl_ca_certs'] == '/path/to/ca.crt'
        assert x.connparams['ssl_certfile'] == '/path/to/client.crt'
        assert x.connparams['ssl_keyfile'] == '/path/to/client.key'

        from redis.connection import SSLConnection
        assert x.connparams['connection_class'] is SSLConnection

    def test_backend_health_check_interval_ssl(self):
        pytest.importorskip('redis')

        self.app.conf.redis_backend_use_ssl = {
            'ssl_cert_reqs': ssl.CERT_REQUIRED,
            'ssl_ca_certs': '/path/to/ca.crt',
            'ssl_certfile': '/path/to/client.crt',
            'ssl_keyfile': '/path/to/client.key',
        }
        self.app.conf.redis_backend_health_check_interval = 10
        x = self.Backend(
            'rediss://:bosco@vandelay.com:123//1', app=self.app,
        )
        assert x.connparams
        assert x.connparams['host'] == 'vandelay.com'
        assert x.connparams['db'] == 1
        assert x.connparams['port'] == 123
        assert x.connparams['password'] == 'bosco'
        assert x.connparams['health_check_interval'] == 10

        from redis.connection import SSLConnection
        assert x.connparams['connection_class'] is SSLConnection

    def test_backend_health_check_interval(self):
        pytest.importorskip('redis')

        self.app.conf.redis_backend_health_check_interval = 10
        x = self.Backend(
            'redis://vandelay.com:123//1', app=self.app,
        )
        assert x.connparams
        assert x.connparams['host'] == 'vandelay.com'
        assert x.connparams['db'] == 1
        assert x.connparams['port'] == 123
        assert x.connparams['health_check_interval'] == 10

    def test_backend_health_check_interval_not_set(self):
        pytest.importorskip('redis')

        x = self.Backend(
            'redis://vandelay.com:123//1', app=self.app,
        )
        assert x.connparams
        assert x.connparams['host'] == 'vandelay.com'
        assert x.connparams['db'] == 1
        assert x.connparams['port'] == 123
        assert "health_check_interval" not in x.connparams

    def test_backend_redis_client_name(self):
        pytest.importorskip('redis')

        self.app.conf.redis_client_name = 'celery-worker'
        x = self.Backend(
            'redis://vandelay.com:123//1', app=self.app,
        )
        assert x.connparams
        assert x.connparams['host'] == 'vandelay.com'
        assert x.connparams['db'] == 1
        assert x.connparams['port'] == 123
        assert x.connparams['client_name'] == 'celery-worker'

    def test_backend_redis_client_name_not_set(self):
        pytest.importorskip('redis')

        x = self.Backend(
            'redis://vandelay.com:123//1', app=self.app,
        )
        assert x.connparams
        assert x.connparams['host'] == 'vandelay.com'
        assert x.connparams['db'] == 1
        assert x.connparams['port'] == 123
        assert x.connparams['client_name'] is None

    @pytest.mark.parametrize('cert_str', [
        "required",
        "CERT_REQUIRED",
    ])
    def test_backend_ssl_certreq_str(self, cert_str):
        pytest.importorskip('redis')

        self.app.conf.redis_backend_use_ssl = {
            'ssl_cert_reqs': cert_str,
            'ssl_ca_certs': '/path/to/ca.crt',
            'ssl_certfile': '/path/to/client.crt',
            'ssl_keyfile': '/path/to/client.key',
        }
        self.app.conf.redis_socket_timeout = 30.0
        self.app.conf.redis_socket_connect_timeout = 100.0
        x = self.Backend(
            'rediss://:bosco@vandelay.com:123//1', app=self.app,
        )
        assert x.connparams
        assert x.connparams['host'] == 'vandelay.com'
        assert x.connparams['db'] == 1
        assert x.connparams['port'] == 123
        assert x.connparams['password'] == 'bosco'
        assert x.connparams['socket_timeout'] == 30.0
        assert x.connparams['socket_connect_timeout'] == 100.0
        assert x.connparams['ssl_cert_reqs'] == ssl.CERT_REQUIRED
        assert x.connparams['ssl_ca_certs'] == '/path/to/ca.crt'
        assert x.connparams['ssl_certfile'] == '/path/to/client.crt'
        assert x.connparams['ssl_keyfile'] == '/path/to/client.key'

        from redis.connection import SSLConnection
        assert x.connparams['connection_class'] is SSLConnection

    @pytest.mark.parametrize('cert_str', [
        "required",
        "CERT_REQUIRED",
    ])
    def test_backend_ssl_url(self, cert_str):
        pytest.importorskip('redis')

        self.app.conf.redis_socket_timeout = 30.0
        self.app.conf.redis_socket_connect_timeout = 100.0
        x = self.Backend(
            'rediss://:bosco@vandelay.com:123//1?ssl_cert_reqs=%s' % cert_str,
            app=self.app,
        )
        assert x.connparams
        assert x.connparams['host'] == 'vandelay.com'
        assert x.connparams['db'] == 1
        assert x.connparams['port'] == 123
        assert x.connparams['password'] == 'bosco'
        assert x.connparams['socket_timeout'] == 30.0
        assert x.connparams['socket_connect_timeout'] == 100.0
        assert x.connparams['ssl_cert_reqs'] == ssl.CERT_REQUIRED

        from redis.connection import SSLConnection
        assert x.connparams['connection_class'] is SSLConnection

    @pytest.mark.parametrize('cert_str', [
        "none",
        "CERT_NONE",
    ])
    def test_backend_ssl_url_options(self, cert_str):
        pytest.importorskip('redis')

        x = self.Backend(
            (
                'rediss://:bosco@vandelay.com:123//1'
                '?ssl_cert_reqs={cert_str}'
                '&ssl_ca_certs=%2Fvar%2Fssl%2Fmyca.pem'
                '&ssl_certfile=%2Fvar%2Fssl%2Fredis-server-cert.pem'
                '&ssl_keyfile=%2Fvar%2Fssl%2Fprivate%2Fworker-key.pem'
            ).format(cert_str=cert_str),
            app=self.app,
        )
        assert x.connparams
        assert x.connparams['host'] == 'vandelay.com'
        assert x.connparams['db'] == 1
        assert x.connparams['port'] == 123
        assert x.connparams['password'] == 'bosco'
        assert x.connparams['ssl_cert_reqs'] == ssl.CERT_NONE
        assert x.connparams['ssl_ca_certs'] == '/var/ssl/myca.pem'
        assert x.connparams['ssl_certfile'] == '/var/ssl/redis-server-cert.pem'
        assert x.connparams['ssl_keyfile'] == '/var/ssl/private/worker-key.pem'

    @pytest.mark.parametrize('cert_str', [
        "optional",
        "CERT_OPTIONAL",
    ])
    def test_backend_ssl_url_cert_none(self, cert_str):
        pytest.importorskip('redis')

        x = self.Backend(
            'rediss://:bosco@vandelay.com:123//1?ssl_cert_reqs=%s' % cert_str,
            app=self.app,
        )
        assert x.connparams
        assert x.connparams['host'] == 'vandelay.com'
        assert x.connparams['db'] == 1
        assert x.connparams['port'] == 123
        assert x.connparams['ssl_cert_reqs'] == ssl.CERT_OPTIONAL

        from redis.connection import SSLConnection
        assert x.connparams['connection_class'] is SSLConnection

    @pytest.mark.parametrize("uri", [
        'rediss://:bosco@vandelay.com:123//1?ssl_cert_reqs=CERT_KITTY_CATS',
        'rediss://:bosco@vandelay.com:123//1'
    ])
    def test_backend_ssl_url_invalid(self, uri):
        pytest.importorskip('redis')

        with pytest.raises(ValueError):
            self.Backend(
                uri,
                app=self.app,
            )

    def test_conf_raises_KeyError(self):
        self.app.conf = AttributeDict({
            'result_serializer': 'json',
            'result_cache_max': 1,
            'result_expires': None,
            'accept_content': ['json'],
            'result_accept_content': ['json'],
        })
        self.Backend(app=self.app)

    @patch('celery.backends.redis.logger')
    def test_on_connection_error(self, logger):
        intervals = iter([10, 20, 30])
        exc = KeyError()
        assert self.b.on_connection_error(None, exc, intervals, 1) == 10
        logger.error.assert_called_with(
            self.E_LOST, 1, 'Inf', 'in 10.00 seconds')
        assert self.b.on_connection_error(10, exc, intervals, 2) == 20
        logger.error.assert_called_with(self.E_LOST, 2, 10, 'in 20.00 seconds')
        assert self.b.on_connection_error(10, exc, intervals, 3) == 30
        logger.error.assert_called_with(self.E_LOST, 3, 10, 'in 30.00 seconds')

    @patch('celery.backends.redis.retry_over_time')
    def test_retry_policy_conf(self, retry_over_time):
        self.app.conf.result_backend_transport_options = dict(
            retry_policy=dict(
                max_retries=2,
                interval_start=0,
                interval_step=0.01,
            ),
        )
        b = self.Backend(app=self.app)

        def fn():
            return 1

        # We don't want to re-test retry_over_time, just check we called it
        # with the expected args
        b.ensure(fn, (),)

        retry_over_time.assert_called_with(
            fn, b.connection_errors, (), {}, ANY,
            max_retries=2, interval_start=0, interval_step=0.01, interval_max=1
        )

    def test_exception_safe_to_retry(self):
        b = self.Backend(app=self.app)
        assert not b.exception_safe_to_retry(Exception("failed"))
        assert not b.exception_safe_to_retry(BaseException("failed"))
        assert not b.exception_safe_to_retry(exceptions.RedisError("redis error"))
        assert b.exception_safe_to_retry(exceptions.ConnectionError("service unavailable"))
        assert b.exception_safe_to_retry(exceptions.TimeoutError("timeout"))

    def test_incr(self):
        self.b.client = Mock(name='client')
        self.b.incr('foo')
        self.b.client.incr.assert_called_with('foo')

    def test_expire(self):
        self.b.client = Mock(name='client')
        self.b.expire('foo', 300)
        self.b.client.expire.assert_called_with('foo', 300)

    def test_apply_chord(self, unlock='celery.chord_unlock'):
        self.app.tasks[unlock] = Mock()
        header_result_args = (
            uuid(),
            [self.app.AsyncResult(x) for x in range(3)],
        )
        self.b.apply_chord(header_result_args, None)
        assert self.app.tasks[unlock].apply_async.call_count == 0

    def test_unpack_chord_result(self):
        self.b.exception_to_python = Mock(name='etp')
        decode = Mock(name='decode')
        exc = KeyError()
        tup = decode.return_value = (1, 'id1', states.FAILURE, exc)
        with pytest.raises(ChordError):
            self.b._unpack_chord_result(tup, decode)
        decode.assert_called_with(tup)
        self.b.exception_to_python.assert_called_with(exc)

        exc = ValueError()
        tup = decode.return_value = (2, 'id2', states.RETRY, exc)
        ret = self.b._unpack_chord_result(tup, decode)
        self.b.exception_to_python.assert_called_with(exc)
        assert ret is self.b.exception_to_python()

    def test_on_chord_part_return_no_gid_or_tid(self):
        request = Mock(name='request')
        request.id = request.group = request.group_index = None
        assert self.b.on_chord_part_return(request, 'SUCCESS', 10) is None

    def test_ConnectionPool(self):
        self.b.redis = Mock(name='redis')
        assert self.b._ConnectionPool is None
        assert self.b.ConnectionPool is self.b.redis.ConnectionPool
        assert self.b.ConnectionPool is self.b.redis.ConnectionPool

    def test_expires_defaults_to_config(self):
        self.app.conf.result_expires = 10
        b = self.Backend(expires=None, app=self.app)
        assert b.expires == 10

    def test_expires_is_int(self):
        b = self.Backend(expires=48, app=self.app)
        assert b.expires == 48

    def test_add_to_chord(self):
        b = self.Backend('redis://', app=self.app)
        gid = uuid()
        b.add_to_chord(gid, 'sig')
        b.client.incr.assert_called_with(b.get_key_for_group(gid, '.t'), 1)

    def test_set_chord_size(self):
        b = self.Backend('redis://', app=self.app)
        gid = uuid()
        b.set_chord_size(gid, 10)
        b.client.set.assert_called_with(b.get_key_for_group(gid, '.s'), 10)

    def test_expires_is_None(self):
        b = self.Backend(expires=None, app=self.app)
        assert b.expires == self.app.conf.result_expires.total_seconds()

    def test_expires_is_timedelta(self):
        b = self.Backend(expires=timedelta(minutes=1), app=self.app)
        assert b.expires == 60

    def test_mget(self):
        assert self.b.mget(['a', 'b', 'c'])
        self.b.client.mget.assert_called_with(['a', 'b', 'c'])

    def test_set_no_expire(self):
        self.b.expires = None
        self.b._set_with_state('foo', 'bar', states.SUCCESS)

    def test_process_cleanup(self):
        self.b.process_cleanup()

    def test_get_set_forget(self):
        tid = uuid()
        self.b.store_result(tid, 42, states.SUCCESS)
        assert self.b.get_state(tid) == states.SUCCESS
        assert self.b.get_result(tid) == 42
        self.b.forget(tid)
        assert self.b.get_state(tid) == states.PENDING

    def test_set_expires(self):
        self.b = self.Backend(expires=512, app=self.app)
        tid = uuid()
        key = self.b.get_key_for_task(tid)
        self.b.store_result(tid, 42, states.SUCCESS)
        self.b.client.expire.assert_called_with(
            key, 512,
        )

    def test_set_raises_error_on_large_value(self):
        with pytest.raises(BackendStoreError):
            self.b.set('key', 'x' * (self.b._MAX_STR_VALUE_SIZE + 1))


class test_RedisBackend_chords_simple(basetest_RedisBackend):
    @pytest.fixture(scope="class", autouse=True)
    def simple_header_result(self):
        with patch(
            "celery.result.GroupResult.restore", return_value=None,
        ) as p:
            yield p

    def test_on_chord_part_return(self):
        tasks = [self.create_task(i) for i in range(10)]
        random.shuffle(tasks)

        for i in range(10):
            self.b.on_chord_part_return(tasks[i].request, states.SUCCESS, i)
            assert self.b.client.zadd.call_count
            self.b.client.zadd.reset_mock()
        assert self.b.client.zrangebyscore.call_count
        jkey = self.b.get_key_for_group('group_id', '.j')
        tkey = self.b.get_key_for_group('group_id', '.t')
        skey = self.b.get_key_for_group('group_id', '.s')
        self.b.client.delete.assert_has_calls([call(jkey), call(tkey), call(skey)])
        self.b.client.expire.assert_has_calls([
            call(jkey, 86400), call(tkey, 86400), call(skey, 86400),
        ])

    def test_on_chord_part_return__unordered(self):
        self.app.conf.result_backend_transport_options = dict(
            result_chord_ordered=False,
        )

        tasks = [self.create_task(i) for i in range(10)]
        random.shuffle(tasks)

        for i in range(10):
            self.b.on_chord_part_return(tasks[i].request, states.SUCCESS, i)
            assert self.b.client.rpush.call_count
            self.b.client.rpush.reset_mock()
        assert self.b.client.lrange.call_count
        jkey = self.b.get_key_for_group('group_id', '.j')
        tkey = self.b.get_key_for_group('group_id', '.t')
        self.b.client.delete.assert_has_calls([call(jkey), call(tkey)])
        self.b.client.expire.assert_has_calls([
            call(jkey, 86400), call(tkey, 86400),
        ])

    def test_on_chord_part_return__ordered(self):
        self.app.conf.result_backend_transport_options = dict(
            result_chord_ordered=True,
        )

        tasks = [self.create_task(i) for i in range(10)]
        random.shuffle(tasks)

        for i in range(10):
            self.b.on_chord_part_return(tasks[i].request, states.SUCCESS, i)
            assert self.b.client.zadd.call_count
            self.b.client.zadd.reset_mock()
        assert self.b.client.zrangebyscore.call_count
        jkey = self.b.get_key_for_group('group_id', '.j')
        tkey = self.b.get_key_for_group('group_id', '.t')
        self.b.client.delete.assert_has_calls([call(jkey), call(tkey)])
        self.b.client.expire.assert_has_calls([
            call(jkey, 86400), call(tkey, 86400),
        ])

    def test_on_chord_part_return_no_expiry(self):
        old_expires = self.b.expires
        self.b.expires = None
        tasks = [self.create_task(i) for i in range(10)]
        self.b.set_chord_size('group_id', 10)

        for i in range(10):
            self.b.on_chord_part_return(tasks[i].request, states.SUCCESS, i)
            assert self.b.client.zadd.call_count
            self.b.client.zadd.reset_mock()
        assert self.b.client.zrangebyscore.call_count
        jkey = self.b.get_key_for_group('group_id', '.j')
        tkey = self.b.get_key_for_group('group_id', '.t')
        self.b.client.delete.assert_has_calls([call(jkey), call(tkey)])
        self.b.client.expire.assert_not_called()

        self.b.expires = old_expires

    def test_on_chord_part_return_expire_set_to_zero(self):
        old_expires = self.b.expires
        self.b.expires = 0
        tasks = [self.create_task(i) for i in range(10)]

        for i in range(10):
            self.b.on_chord_part_return(tasks[i].request, states.SUCCESS, i)
            assert self.b.client.zadd.call_count
            self.b.client.zadd.reset_mock()
        assert self.b.client.zrangebyscore.call_count
        jkey = self.b.get_key_for_group('group_id', '.j')
        tkey = self.b.get_key_for_group('group_id', '.t')
        self.b.client.delete.assert_has_calls([call(jkey), call(tkey)])
        self.b.client.expire.assert_not_called()

        self.b.expires = old_expires

    def test_on_chord_part_return_no_expiry__unordered(self):
        self.app.conf.result_backend_transport_options = dict(
            result_chord_ordered=False,
        )

        old_expires = self.b.expires
        self.b.expires = None
        tasks = [self.create_task(i) for i in range(10)]

        for i in range(10):
            self.b.on_chord_part_return(tasks[i].request, states.SUCCESS, i)
            assert self.b.client.rpush.call_count
            self.b.client.rpush.reset_mock()
        assert self.b.client.lrange.call_count
        jkey = self.b.get_key_for_group('group_id', '.j')
        tkey = self.b.get_key_for_group('group_id', '.t')
        self.b.client.delete.assert_has_calls([call(jkey), call(tkey)])
        self.b.client.expire.assert_not_called()

        self.b.expires = old_expires

    def test_on_chord_part_return_no_expiry__ordered(self):
        self.app.conf.result_backend_transport_options = dict(
            result_chord_ordered=True,
        )

        old_expires = self.b.expires
        self.b.expires = None
        tasks = [self.create_task(i) for i in range(10)]

        for i in range(10):
            self.b.on_chord_part_return(tasks[i].request, states.SUCCESS, i)
            assert self.b.client.zadd.call_count
            self.b.client.zadd.reset_mock()
        assert self.b.client.zrangebyscore.call_count
        jkey = self.b.get_key_for_group('group_id', '.j')
        tkey = self.b.get_key_for_group('group_id', '.t')
        self.b.client.delete.assert_has_calls([call(jkey), call(tkey)])
        self.b.client.expire.assert_not_called()

        self.b.expires = old_expires

    def test_on_chord_part_return__success(self):
        with self.chord_context(2) as (_, request, callback):
            self.b.on_chord_part_return(request, states.SUCCESS, 10)
            callback.delay.assert_not_called()
            self.b.on_chord_part_return(request, states.SUCCESS, 20)
            callback.delay.assert_called_with([10, 20])

    def test_on_chord_part_return__success__unordered(self):
        self.app.conf.result_backend_transport_options = dict(
            result_chord_ordered=False,
        )

        with self.chord_context(2) as (_, request, callback):
            self.b.on_chord_part_return(request, states.SUCCESS, 10)
            callback.delay.assert_not_called()
            self.b.on_chord_part_return(request, states.SUCCESS, 20)
            callback.delay.assert_called_with([10, 20])

    def test_on_chord_part_return__success__ordered(self):
        self.app.conf.result_backend_transport_options = dict(
            result_chord_ordered=True,
        )

        with self.chord_context(2) as (_, request, callback):
            self.b.on_chord_part_return(request, states.SUCCESS, 10)
            callback.delay.assert_not_called()
            self.b.on_chord_part_return(request, states.SUCCESS, 20)
            callback.delay.assert_called_with([10, 20])

    def test_on_chord_part_return__callback_raises(self):
        with self.chord_context(1) as (_, request, callback):
            callback.delay.side_effect = KeyError(10)
            task = self.app._tasks['add'] = Mock(name='add_task')
            self.b.on_chord_part_return(request, states.SUCCESS, 10)
            task.backend.fail_from_current_stack.assert_called_with(
                callback.id, exc=ANY,
            )

    def test_on_chord_part_return__callback_raises__unordered(self):
        self.app.conf.result_backend_transport_options = dict(
            result_chord_ordered=False,
        )

        with self.chord_context(1) as (_, request, callback):
            callback.delay.side_effect = KeyError(10)
            task = self.app._tasks['add'] = Mock(name='add_task')
            self.b.on_chord_part_return(request, states.SUCCESS, 10)
            task.backend.fail_from_current_stack.assert_called_with(
                callback.id, exc=ANY,
            )

    def test_on_chord_part_return__callback_raises__ordered(self):
        self.app.conf.result_backend_transport_options = dict(
            result_chord_ordered=True,
        )

        with self.chord_context(1) as (_, request, callback):
            callback.delay.side_effect = KeyError(10)
            task = self.app._tasks['add'] = Mock(name='add_task')
            self.b.on_chord_part_return(request, states.SUCCESS, 10)
            task.backend.fail_from_current_stack.assert_called_with(
                callback.id, exc=ANY,
            )

    def test_on_chord_part_return__ChordError(self):
        with self.chord_context(1) as (_, request, callback):
            self.b.client.pipeline = ContextMock()
            raise_on_second_call(self.b.client.pipeline, ChordError())
            self.b.client.pipeline.return_value.zadd().zcount().get().get().expire(
            ).expire().expire().execute.return_value = (1, 1, 0, b'1', 4, 5, 6)
            task = self.app._tasks['add'] = Mock(name='add_task')
            self.b.on_chord_part_return(request, states.SUCCESS, 10)
            task.backend.fail_from_current_stack.assert_called_with(
                callback.id, exc=ANY,
            )

    def test_on_chord_part_return__ChordError__unordered(self):
        self.app.conf.result_backend_transport_options = dict(
            result_chord_ordered=False,
        )

        with self.chord_context(1) as (_, request, callback):
            self.b.client.pipeline = ContextMock()
            raise_on_second_call(self.b.client.pipeline, ChordError())
            self.b.client.pipeline.return_value.rpush().llen().get().get().expire(
            ).expire().expire().execute.return_value = (1, 1, 0, b'1', 4, 5, 6)
            task = self.app._tasks['add'] = Mock(name='add_task')
            self.b.on_chord_part_return(request, states.SUCCESS, 10)
            task.backend.fail_from_current_stack.assert_called_with(
                callback.id, exc=ANY,
            )

    def test_on_chord_part_return__ChordError__ordered(self):
        self.app.conf.result_backend_transport_options = dict(
            result_chord_ordered=True,
        )

        with self.chord_context(1) as (_, request, callback):
            self.b.client.pipeline = ContextMock()
            raise_on_second_call(self.b.client.pipeline, ChordError())
            self.b.client.pipeline.return_value.zadd().zcount().get().get().expire(
            ).expire().expire().execute.return_value = (1, 1, 0, b'1', 4, 5, 6)
            task = self.app._tasks['add'] = Mock(name='add_task')
            self.b.on_chord_part_return(request, states.SUCCESS, 10)
            task.backend.fail_from_current_stack.assert_called_with(
                callback.id, exc=ANY,
            )

    def test_on_chord_part_return__other_error(self):
        with self.chord_context(1) as (_, request, callback):
            self.b.client.pipeline = ContextMock()
            raise_on_second_call(self.b.client.pipeline, RuntimeError())
            self.b.client.pipeline.return_value.zadd().zcount().get().get().expire(
            ).expire().expire().execute.return_value = (1, 1, 0, b'1', 4, 5, 6)
            task = self.app._tasks['add'] = Mock(name='add_task')
            self.b.on_chord_part_return(request, states.SUCCESS, 10)
            task.backend.fail_from_current_stack.assert_called_with(
                callback.id, exc=ANY,
            )

    def test_on_chord_part_return__other_error__unordered(self):
        self.app.conf.result_backend_transport_options = dict(
            result_chord_ordered=False,
        )

        with self.chord_context(1) as (_, request, callback):
            self.b.client.pipeline = ContextMock()
            raise_on_second_call(self.b.client.pipeline, RuntimeError())
            self.b.client.pipeline.return_value.rpush().llen().get().get().expire(
            ).expire().expire().execute.return_value = (1, 1, 0, b'1', 4, 5, 6)
            task = self.app._tasks['add'] = Mock(name='add_task')
            self.b.on_chord_part_return(request, states.SUCCESS, 10)
            task.backend.fail_from_current_stack.assert_called_with(
                callback.id, exc=ANY,
            )

    def test_on_chord_part_return__other_error__ordered(self):
        self.app.conf.result_backend_transport_options = dict(
            result_chord_ordered=True,
        )

        with self.chord_context(1) as (_, request, callback):
            self.b.client.pipeline = ContextMock()
            raise_on_second_call(self.b.client.pipeline, RuntimeError())
            self.b.client.pipeline.return_value.zadd().zcount().get().get().expire(
            ).expire().expire().execute.return_value = (1, 1, 0, b'1', 4, 5, 6)
            task = self.app._tasks['add'] = Mock(name='add_task')
            self.b.on_chord_part_return(request, states.SUCCESS, 10)
            task.backend.fail_from_current_stack.assert_called_with(
                callback.id, exc=ANY,
            )


class test_RedisBackend_chords_complex(basetest_RedisBackend):
    @pytest.fixture(scope="function", autouse=True)
    def complex_header_result(self):
        with patch("celery.result.GroupResult.restore") as p:
            yield p

    @pytest.mark.parametrize(['results', 'assert_save_called'], [
        # No results in the header at all - won't call `save()`
        (tuple(), False),
        # Simple results in the header - won't call `save()`
        ((AsyncResult("foo"), ), False),
        # Many simple results in the header - won't call `save()`
        ((AsyncResult("foo"), ) * 42, False),
        # A single complex result in the header - will call `save()`
        ((GroupResult("foo", []),), True),
        # Many complex results in the header - will call `save()`
        ((GroupResult("foo"), ) * 42, True),
        # Mixed simple and complex results in the header - will call `save()`
        (itertools.islice(
            itertools.cycle((
                AsyncResult("foo"), GroupResult("foo"),
            )), 42,
        ), True),
    ])
    def test_apply_chord_complex_header(self, results, assert_save_called):
        mock_group_result = Mock()
        mock_group_result.return_value.results = results
        self.app.GroupResult = mock_group_result
        header_result_args = ("gid11", results)
        self.b.apply_chord(header_result_args, None)
        if assert_save_called:
            mock_group_result.return_value.save.assert_called_once_with(backend=self.b)
        else:
            mock_group_result.return_value.save.assert_not_called()

    def test_on_chord_part_return_timeout(self, complex_header_result):
        tasks = [self.create_task(i) for i in range(10)]
        random.shuffle(tasks)
        try:
            self.app.conf.result_chord_join_timeout += 1.0
            for task, result_val in zip(tasks, itertools.cycle((42, ))):
                self.b.on_chord_part_return(
                    task.request, states.SUCCESS, result_val,
                )
        finally:
            self.app.conf.result_chord_join_timeout -= 1.0

        join_func = complex_header_result.return_value.join_native
        join_func.assert_called_once_with(timeout=4.0, propagate=True)

    @pytest.mark.parametrize("supports_native_join", (True, False))
    def test_on_chord_part_return(
        self, complex_header_result, supports_native_join,
    ):
        mock_result_obj = complex_header_result.return_value
        mock_result_obj.supports_native_join = supports_native_join

        tasks = [self.create_task(i) for i in range(10)]
        random.shuffle(tasks)

        with self.chord_context(10) as (tasks, request, callback):
            for task, result_val in zip(tasks, itertools.cycle((42, ))):
                self.b.on_chord_part_return(
                    task.request, states.SUCCESS, result_val,
                )
                # Confirm that `zadd` was called even though we won't end up
                # using the data pushed into the sorted set
                assert self.b.client.zadd.call_count == 1
                self.b.client.zadd.reset_mock()
        # Confirm that neither `zrange` not `lrange` were called
        self.b.client.zrange.assert_not_called()
        self.b.client.lrange.assert_not_called()
        # Confirm that the `GroupResult.restore` mock was called
        complex_header_result.assert_called_once_with(request.group, app=self.b.app)
        # Confirm that the callback was called with the `join()`ed group result
        if supports_native_join:
            expected_join = mock_result_obj.join_native
        else:
            expected_join = mock_result_obj.join
        callback.delay.assert_called_once_with(expected_join())


class test_SentinelBackend:
    def get_backend(self):
        from celery.backends.redis import SentinelBackend

        class _SentinelBackend(SentinelBackend):
            redis = redis
            sentinel = sentinel

        return _SentinelBackend

    def get_E_LOST(self):
        from celery.backends.redis import E_LOST
        return E_LOST

    def setup_method(self):
        self.Backend = self.get_backend()
        self.E_LOST = self.get_E_LOST()
        self.b = self.Backend(app=self.app)

    @pytest.mark.usefixtures('depends_on_current_app')
    def test_reduce(self):
        pytest.importorskip('redis')

        from celery.backends.redis import SentinelBackend
        x = SentinelBackend(app=self.app)
        assert loads(dumps(x))

    def test_no_redis(self):
        self.Backend.redis = None
        with pytest.raises(ImproperlyConfigured):
            self.Backend(app=self.app)

    def test_url(self):
        self.app.conf.redis_socket_timeout = 30.0
        self.app.conf.redis_socket_connect_timeout = 100.0
        x = self.Backend(
            'sentinel://:test@github.com:123/1;'
            'sentinel://:test@github.com:124/1',
            app=self.app,
        )
        assert x.connparams
        assert "host" not in x.connparams
        assert x.connparams['db'] == 1
        assert "port" not in x.connparams
        assert x.connparams['password'] == "test"
        assert len(x.connparams['hosts']) == 2
        expected_hosts = ["github.com", "github.com"]
        found_hosts = [cp['host'] for cp in x.connparams['hosts']]
        assert found_hosts == expected_hosts

        expected_ports = [123, 124]
        found_ports = [cp['port'] for cp in x.connparams['hosts']]
        assert found_ports == expected_ports

        expected_passwords = ["test", "test"]
        found_passwords = [cp['password'] for cp in x.connparams['hosts']]
        assert found_passwords == expected_passwords

        expected_dbs = [1, 1]
        found_dbs = [cp['db'] for cp in x.connparams['hosts']]
        assert found_dbs == expected_dbs

        # By default passwords should be sanitized
        display_url = x.as_uri()
        assert "test" not in display_url
        # We can choose not to sanitize with the `include_password` argument
        unsanitized_display_url = x.as_uri(include_password=True)
        assert unsanitized_display_url == x.url
        # or to explicitly sanitize
        forcibly_sanitized_display_url = x.as_uri(include_password=False)
        assert forcibly_sanitized_display_url == display_url

    def test_get_sentinel_instance(self):
        x = self.Backend(
            'sentinel://:test@github.com:123/1;'
            'sentinel://:test@github.com:124/1',
            app=self.app,
        )
        sentinel_instance = x._get_sentinel_instance(**x.connparams)
        assert sentinel_instance.sentinel_kwargs == {}
        assert sentinel_instance.connection_kwargs['db'] == 1
        assert sentinel_instance.connection_kwargs['password'] == "test"
        assert len(sentinel_instance.sentinels) == 2

    def test_get_pool(self):
        x = self.Backend(
            'sentinel://:test@github.com:123/1;'
            'sentinel://:test@github.com:124/1',
            app=self.app,
        )
        pool = x._get_pool(**x.connparams)
        assert pool

    def test_backend_ssl(self):
        pytest.importorskip('redis')

        from celery.backends.redis import SentinelBackend
        self.app.conf.redis_backend_use_ssl = {
            'ssl_cert_reqs': "CERT_REQUIRED",
            'ssl_ca_certs': '/path/to/ca.crt',
            'ssl_certfile': '/path/to/client.crt',
            'ssl_keyfile': '/path/to/client.key',
        }
        self.app.conf.redis_socket_timeout = 30.0
        self.app.conf.redis_socket_connect_timeout = 100.0
        x = SentinelBackend(
            'sentinel://:bosco@vandelay.com:123//1', app=self.app,
        )
        assert x.connparams
        assert len(x.connparams['hosts']) == 1
        assert x.connparams['hosts'][0]['host'] == 'vandelay.com'
        assert x.connparams['hosts'][0]['db'] == 1
        assert x.connparams['hosts'][0]['port'] == 123
        assert x.connparams['hosts'][0]['password'] == 'bosco'
        assert x.connparams['socket_timeout'] == 30.0
        assert x.connparams['socket_connect_timeout'] == 100.0
        assert x.connparams['ssl_cert_reqs'] == ssl.CERT_REQUIRED
        assert x.connparams['ssl_ca_certs'] == '/path/to/ca.crt'
        assert x.connparams['ssl_certfile'] == '/path/to/client.crt'
        assert x.connparams['ssl_keyfile'] == '/path/to/client.key'

        from celery.backends.redis import SentinelManagedSSLConnection
        assert x.connparams['connection_class'] is SentinelManagedSSLConnection

    def test_url_with_acl_credentials(self):
        x = self.Backend(
            'sentinel://myuser:mypass@github.com:123/1;'
            'sentinel://myuser:mypass@github.com:124/1',
            app=self.app,
        )
        assert x.connparams
        assert "host" not in x.connparams
        assert x.connparams['db'] == 1
        assert "port" not in x.connparams
        assert x.connparams['password'] == "mypass"
        assert x.connparams['username'] == "myuser"
        assert len(x.connparams['hosts']) == 2

        expected_usernames = ["myuser", "myuser"]
        found_usernames = [cp['username'] for cp in x.connparams['hosts']]
        assert found_usernames == expected_usernames

    def test_get_pool_with_acl_credentials(self):
        x = self.Backend(
            'sentinel://myuser:mypass@github.com:123/1;'
            'sentinel://myuser:mypass@github.com:124/1',
            app=self.app,
        )
        with patch.object(x, '_get_sentinel_instance') as mock_get_sentinel:
            mock_sentinel = Mock()
            mock_sentinel.master_for.return_value = Mock(connection_pool=Mock())
            mock_get_sentinel.return_value = mock_sentinel

            x._get_pool(**x.connparams)

            mock_sentinel.master_for.assert_called_once()
            call_kwargs = mock_sentinel.master_for.call_args[1]
            assert call_kwargs.get('username') == 'myuser'
            assert call_kwargs.get('password') == 'mypass'

    def test_get_pool_with_password_only(self):
        x = self.Backend(
            'sentinel://:mypass@github.com:123/1',
            app=self.app,
        )
        with patch.object(x, '_get_sentinel_instance') as mock_get_sentinel:
            mock_sentinel = Mock()
            mock_sentinel.master_for.return_value = Mock(connection_pool=Mock())
            mock_get_sentinel.return_value = mock_sentinel

            x._get_pool(**x.connparams)

            mock_sentinel.master_for.assert_called_once()
            call_kwargs = mock_sentinel.master_for.call_args[1]
            assert 'username' not in call_kwargs
            assert call_kwargs.get('password') == 'mypass'



# Test config for t/integration/test_worker.py

broker_url = 'amqp://guest:guest@foobar:1234//'

# Fail fast for test_run_worker
broker_connection_retry_on_startup = False
broker_connection_retry = False
broker_connection_timeout = 0

worker_log_color = False

worker_redirect_stdouts = False



from __future__ import annotations

import os

import pytest
from pytest_celery import (LOCALSTACK_CREDS, REDIS_CONTAINER_TIMEOUT, REDIS_ENV, REDIS_IMAGE, REDIS_PORTS,
                           CeleryTestSetup, RedisContainer)
from pytest_docker_tools import container, fetch, fxtr

from celery import Celery
from t.smoke.operations.task_termination import TaskTermination
from t.smoke.operations.worker_kill import WorkerKill
from t.smoke.operations.worker_restart import WorkerRestart
from t.smoke.workers.alt import *  # noqa
from t.smoke.workers.dev import *  # noqa
from t.smoke.workers.latest import *  # noqa
from t.smoke.workers.other import *  # noqa


class SmokeTestSetup(CeleryTestSetup):
    def ready(self, *args, **kwargs) -> bool:
        # Force false, false, true
        return super().ready(
            ping=False,
            control=False,
            docker=True,
        )


@pytest.fixture
def celery_setup_cls() -> type[CeleryTestSetup]:  # type: ignore
    return SmokeTestSetup


class SuiteOperations(
    TaskTermination,
    WorkerKill,
    WorkerRestart,
):
    """Optional operations that can be performed with different methods,
    shared across the smoke tests suite.

    Example Usage:
    >>> class test_mysuite(SuiteOperations):
    >>>     def test_something(self):
    >>>         self.prepare_worker_with_conditions()
    >>>         assert condition are met
    """


@pytest.fixture
def default_worker_tasks(default_worker_tasks: set) -> set:
    """Use all of the integration and smoke suites tasks in the smoke tests workers."""
    from t.integration import tasks as integration_tests_tasks
    from t.smoke import tasks as smoke_tests_tasks

    default_worker_tasks.add(integration_tests_tasks)
    default_worker_tasks.add(smoke_tests_tasks)
    return default_worker_tasks


# When using integration tests tasks that requires a Redis instance,
# we use pytest-celery to raise a dedicated Redis container for the smoke tests suite that is configured
# to be used by the integration tests tasks.

redis_command = RedisContainer.command()
redis_command.insert(1, "/usr/local/etc/redis/redis.conf")

redis_image = fetch(repository=REDIS_IMAGE)
redis_test_container: RedisContainer = container(
    image="{redis_image.id}",
    ports=REDIS_PORTS,
    environment=REDIS_ENV,
    network="{default_pytest_celery_network.name}",
    wrapper_class=RedisContainer,
    timeout=REDIS_CONTAINER_TIMEOUT,
    command=redis_command,
    volumes={
        os.path.abspath("t/smoke/redis.conf"): {
            "bind": "/usr/local/etc/redis/redis.conf",
            "mode": "ro",  # Mount as read-only
        }
    },
)


@pytest.fixture(autouse=True)
def set_redis_test_container(redis_test_container: RedisContainer):
    """Configure the Redis test container to be used by the integration tests tasks."""
    # get_redis_connection(): will use these settings in the tests environment
    os.environ["REDIS_HOST"] = "localhost"
    os.environ["REDIS_PORT"] = str(redis_test_container.port)


@pytest.fixture
def default_worker_env(default_worker_env: dict, redis_test_container: RedisContainer) -> dict:
    """Add the Redis connection details to the worker environment."""
    # get_redis_connection(): will use these settings when executing tasks in the worker
    default_worker_env.update(
        {
            "REDIS_HOST": redis_test_container.hostname,
            "REDIS_PORT": 6379,
            **LOCALSTACK_CREDS,
        }
    )
    return default_worker_env


@pytest.fixture(scope="session", autouse=True)
def set_aws_credentials():
    os.environ.update(LOCALSTACK_CREDS)


@pytest.fixture
def default_worker_app(default_worker_app: Celery) -> Celery:
    app = default_worker_app
    if app.conf.broker_url and app.conf.broker_url.startswith("sqs"):
        app.conf.broker_transport_options["region"] = LOCALSTACK_CREDS["AWS_DEFAULT_REGION"]
    return app


# Override the default redis broker container from pytest-celery
default_redis_broker = container(
    image="{default_redis_broker_image}",
    ports=fxtr("default_redis_broker_ports"),
    environment=fxtr("default_redis_broker_env"),
    network="{default_pytest_celery_network.name}",
    wrapper_class=RedisContainer,
    timeout=REDIS_CONTAINER_TIMEOUT,
    command=redis_command,
    volumes={
        os.path.abspath("t/smoke/redis.conf"): {
            "bind": "/usr/local/etc/redis/redis.conf",
            "mode": "ro",  # Mount as read-only
        }
    },
)


# Override the default redis backend container from pytest-celery
default_redis_backend = container(
    image="{default_redis_backend_image}",
    ports=fxtr("default_redis_backend_ports"),
    environment=fxtr("default_redis_backend_env"),
    network="{default_pytest_celery_network.name}",
    wrapper_class=RedisContainer,
    timeout=REDIS_CONTAINER_TIMEOUT,
    command=redis_command,
    volumes={
        os.path.abspath("t/smoke/redis.conf"): {
            "bind": "/usr/local/etc/redis/redis.conf",
            "mode": "ro",  # Mount as read-only
        }
    },
)



from time import sleep

import pytest
from pytest_celery import CeleryTestSetup, CeleryTestWorker, RabbitMQTestBroker

import celery
from celery import Celery
from celery.canvas import chain, group
from t.smoke.conftest import SuiteOperations, WorkerKill, WorkerRestart
from t.smoke.tasks import long_running_task

RESULT_TIMEOUT = 30


def assert_container_exited(worker: CeleryTestWorker, attempts: int = RESULT_TIMEOUT):
    """It might take a few moments for the container to exit after the worker is killed."""
    while attempts:
        worker.container.reload()
        if worker.container.status == "exited":
            break
        attempts -= 1
        sleep(1)

    worker.container.reload()
    assert worker.container.status == "exited"


@pytest.mark.parametrize("method", list(WorkerRestart.Method))
class test_worker_restart(SuiteOperations):
    @pytest.fixture
    def default_worker_app(self, default_worker_app: Celery) -> Celery:
        app = default_worker_app
        app.conf.worker_pool_restarts = True
        app.conf.task_acks_late = True
        return app

    def test_restart_during_task_execution(
        self,
        celery_setup: CeleryTestSetup,
        method: WorkerRestart.Method,
    ):
        queue = celery_setup.worker.worker_queue
        sig = long_running_task.si(5, verbose=True).set(queue=queue)
        res = sig.delay()
        self.restart_worker(celery_setup.worker, method)
        assert res.get(RESULT_TIMEOUT) is True

    def test_restart_between_task_execution(
        self,
        celery_setup: CeleryTestSetup,
        method: WorkerRestart.Method,
    ):
        # We use freeze() to control the order of execution for the restart operation
        queue = celery_setup.worker.worker_queue
        first = long_running_task.si(5, verbose=True).set(queue=queue)
        first_res = first.freeze()
        second = long_running_task.si(5, verbose=True).set(queue=queue)
        second_res = second.freeze()
        sig = chain(first, second)
        sig.delay()
        assert first_res.get(RESULT_TIMEOUT) is True
        self.restart_worker(celery_setup.worker, method)
        assert second_res.get(RESULT_TIMEOUT) is True


class test_worker_shutdown(SuiteOperations):
    @pytest.fixture
    def default_worker_app(self, default_worker_app: Celery) -> Celery:
        app = default_worker_app
        app.conf.task_acks_late = True
        return app

    def test_warm_shutdown(self, celery_setup: CeleryTestSetup):
        queue = celery_setup.worker.worker_queue
        worker = celery_setup.worker
        sig = long_running_task.si(5, verbose=True).set(queue=queue)
        res = sig.delay()

        worker.assert_log_exists("Starting long running task")
        self.kill_worker(worker, WorkerKill.Method.SIGTERM)
        worker.assert_log_exists("worker: Warm shutdown (MainProcess)")

        assert_container_exited(worker)
        assert res.get(RESULT_TIMEOUT)

    def test_multiple_warm_shutdown_does_nothing(self, celery_setup: CeleryTestSetup):
        queue = celery_setup.worker.worker_queue
        worker = celery_setup.worker
        sig = long_running_task.si(5, verbose=True).set(queue=queue)
        res = sig.delay()

        worker.assert_log_exists("Starting long running task")
        for _ in range(3):
            self.kill_worker(worker, WorkerKill.Method.SIGTERM)

        assert_container_exited(worker)
        assert res.get(RESULT_TIMEOUT)

    def test_cold_shutdown(self, celery_setup: CeleryTestSetup):
        queue = celery_setup.worker.worker_queue
        worker = celery_setup.worker
        sig = long_running_task.si(5, verbose=True).set(queue=queue)
        res = sig.delay()

        worker.assert_log_exists("Starting long running task")
        self.kill_worker(worker, WorkerKill.Method.SIGQUIT)
        worker.assert_log_exists("worker: Cold shutdown (MainProcess)")
        worker.assert_log_does_not_exist(f"long_running_task[{res.id}] succeeded", timeout=10)

        assert_container_exited(worker)

        with pytest.raises(celery.exceptions.TimeoutError):
            res.get(timeout=5)

    def test_hard_shutdown_from_warm(self, celery_setup: CeleryTestSetup):
        queue = celery_setup.worker.worker_queue
        worker = celery_setup.worker
        sig = long_running_task.si(420, verbose=True).set(queue=queue)
        sig.delay()

        worker.assert_log_exists("Starting long running task")
        self.kill_worker(worker, WorkerKill.Method.SIGTERM)
        self.kill_worker(worker, WorkerKill.Method.SIGQUIT)
        self.kill_worker(worker, WorkerKill.Method.SIGQUIT)

        worker.assert_log_exists("worker: Warm shutdown (MainProcess)")
        worker.assert_log_exists("worker: Cold shutdown (MainProcess)")

        assert_container_exited(worker)

    def test_hard_shutdown_from_cold(self, celery_setup: CeleryTestSetup):
        queue = celery_setup.worker.worker_queue
        worker = celery_setup.worker
        sig = long_running_task.si(420, verbose=True).set(queue=queue)
        sig.delay()

        worker.assert_log_exists("Starting long running task")
        self.kill_worker(worker, WorkerKill.Method.SIGQUIT)
        self.kill_worker(worker, WorkerKill.Method.SIGQUIT)

        worker.assert_log_exists("worker: Cold shutdown (MainProcess)")

        assert_container_exited(worker)

    class test_REMAP_SIGTERM(SuiteOperations):
        @pytest.fixture
        def default_worker_env(self, default_worker_env: dict) -> dict:
            default_worker_env.update({"REMAP_SIGTERM": "SIGQUIT"})
            return default_worker_env

        def test_cold_shutdown(self, celery_setup: CeleryTestSetup):
            queue = celery_setup.worker.worker_queue
            worker = celery_setup.worker
            sig = long_running_task.si(5, verbose=True).set(queue=queue)
            res = sig.delay()

            worker.assert_log_exists("Starting long running task")
            self.kill_worker(worker, WorkerKill.Method.SIGTERM)
            worker.assert_log_exists("worker: Cold shutdown (MainProcess)")
            worker.assert_log_does_not_exist(f"long_running_task[{res.id}] succeeded", timeout=10)

            assert_container_exited(worker)

        def test_hard_shutdown_from_cold(self, celery_setup: CeleryTestSetup):
            queue = celery_setup.worker.worker_queue
            worker = celery_setup.worker
            sig = long_running_task.si(420, verbose=True).set(queue=queue)
            sig.delay()

            worker.assert_log_exists("Starting long running task")
            self.kill_worker(worker, WorkerKill.Method.SIGTERM)
            self.kill_worker(worker, WorkerKill.Method.SIGTERM)

            worker.assert_log_exists("worker: Cold shutdown (MainProcess)")

            assert_container_exited(worker)

    class test_worker_soft_shutdown_timeout(SuiteOperations):
        @pytest.fixture
        def default_worker_app(self, default_worker_app: Celery) -> Celery:
            app = default_worker_app
            app.conf.worker_soft_shutdown_timeout = 10
            return app

        def test_soft_shutdown(self, celery_setup: CeleryTestSetup):
            app = celery_setup.app
            queue = celery_setup.worker.worker_queue
            worker = celery_setup.worker
            sig = long_running_task.si(5, verbose=True).set(queue=queue)
            res = sig.delay()

            worker.assert_log_exists("Starting long running task")
            self.kill_worker(worker, WorkerKill.Method.SIGQUIT)
            worker.assert_log_exists(
                f"Initiating Soft Shutdown, terminating in {app.conf.worker_soft_shutdown_timeout} seconds",
                timeout=5,
            )
            worker.assert_log_exists("worker: Cold shutdown (MainProcess)")

            assert_container_exited(worker)
            assert res.get(RESULT_TIMEOUT)

        def test_hard_shutdown_from_soft(self, celery_setup: CeleryTestSetup):
            queue = celery_setup.worker.worker_queue
            worker = celery_setup.worker
            sig = long_running_task.si(420, verbose=True).set(queue=queue)
            sig.delay()

            worker.assert_log_exists("Starting long running task")
            self.kill_worker(worker, WorkerKill.Method.SIGQUIT)
            self.kill_worker(worker, WorkerKill.Method.SIGQUIT)
            worker.assert_log_exists("Waiting gracefully for cold shutdown to complete...")
            worker.assert_log_exists("worker: Cold shutdown (MainProcess)")
            self.kill_worker(worker, WorkerKill.Method.SIGQUIT)

            assert_container_exited(worker)

        def test_task_completes_during_soft_shutdown(self, celery_setup: CeleryTestSetup):
            app = celery_setup.app
            queue = celery_setup.worker.worker_queue
            worker = celery_setup.worker

            task_duration = app.conf.worker_soft_shutdown_timeout // 2
            sig = long_running_task.si(task_duration, verbose=True).set(queue=queue)
            res = sig.delay()

            worker.assert_log_exists("Starting long running task")
            self.kill_worker(worker, WorkerKill.Method.SIGQUIT)
            worker.assert_log_exists(
                f"Initiating Soft Shutdown, terminating in {app.conf.worker_soft_shutdown_timeout} seconds"
            )
            worker.assert_log_exists("worker: Cold shutdown (MainProcess)")

            assert_container_exited(worker)
            assert res.get(RESULT_TIMEOUT)
            assert res.state == 'SUCCESS'

        class test_time_limit(SuiteOperations):
            @pytest.fixture
            def default_worker_app(self, default_worker_app: Celery) -> Celery:
                app = default_worker_app
                app.conf.worker_soft_shutdown_timeout = 16
                app.conf.task_time_limit = 15
                app.conf.task_soft_time_limit = 10
                return app

            def test_task_completes_during_soft_shutdown_with_time_limit(self, celery_setup: CeleryTestSetup):
                app = celery_setup.app
                queue = celery_setup.worker.worker_queue
                worker = celery_setup.worker

                task_duration = 8
                sig = long_running_task.si(task_duration, verbose=True).set(queue=queue)
                res = sig.delay()

                worker.assert_log_exists("Starting long running task")
                self.kill_worker(worker, WorkerKill.Method.SIGQUIT)

                worker.assert_log_exists(
                    f"Initiating Soft Shutdown, terminating in {app.conf.worker_soft_shutdown_timeout} seconds"
                )

                worker.assert_log_exists("Finished long running task")
                worker.assert_log_exists(f"long_running_task[{res.id}] succeeded")

                worker.assert_log_does_not_exist(
                    f"Task handler raised error: TimeLimitExceeded({app.conf.task_time_limit})",
                    timeout=RESULT_TIMEOUT,
                )
                worker.assert_log_does_not_exist(
                    f"Hard time limit ({app.conf.task_time_limit}s) exceeded for {long_running_task.name}[{res.id}]",
                    timeout=RESULT_TIMEOUT,
                )
                worker.assert_log_exists("worker: Cold shutdown (MainProcess)")

                assert_container_exited(worker)
                assert res.get(RESULT_TIMEOUT)
                assert res.state == 'SUCCESS'

        class test_REMAP_SIGTERM(SuiteOperations):
            @pytest.fixture
            def default_worker_env(self, default_worker_env: dict) -> dict:
                default_worker_env.update({"REMAP_SIGTERM": "SIGQUIT"})
                return default_worker_env

            def test_soft_shutdown(self, celery_setup: CeleryTestSetup):
                app = celery_setup.app
                queue = celery_setup.worker.worker_queue
                worker = celery_setup.worker
                sig = long_running_task.si(5, verbose=True).set(queue=queue)
                res = sig.delay()

                worker.assert_log_exists("Starting long running task")
                self.kill_worker(worker, WorkerKill.Method.SIGTERM)
                worker.assert_log_exists(
                    f"Initiating Soft Shutdown, terminating in {app.conf.worker_soft_shutdown_timeout} seconds"
                )
                worker.assert_log_exists("worker: Cold shutdown (MainProcess)")

                assert_container_exited(worker)
                assert res.get(RESULT_TIMEOUT)

            def test_hard_shutdown_from_soft(self, celery_setup: CeleryTestSetup):
                queue = celery_setup.worker.worker_queue
                worker = celery_setup.worker
                sig = long_running_task.si(420, verbose=True).set(queue=queue)
                sig.delay()

                worker.assert_log_exists("Starting long running task")
                self.kill_worker(worker, WorkerKill.Method.SIGTERM)
                self.kill_worker(worker, WorkerKill.Method.SIGTERM)
                worker.assert_log_exists("Waiting gracefully for cold shutdown to complete...")
                worker.assert_log_exists("worker: Cold shutdown (MainProcess)", timeout=5)
                self.kill_worker(worker, WorkerKill.Method.SIGTERM)

                assert_container_exited(worker)

        class test_reset_visibility_timeout(SuiteOperations):
            @pytest.fixture
            def default_worker_app(self, default_worker_app: Celery) -> Celery:
                app = default_worker_app
                app.conf.prefetch_multiplier = 2
                app.conf.worker_concurrency = 10
                app.conf.visibility_timeout = 3600  # 1 hour
                app.conf.broker_transport_options = {
                    "visibility_timeout": app.conf.visibility_timeout,
                    "polling_interval": 1,
                }
                app.conf.result_backend_transport_options = {
                    "visibility_timeout": app.conf.visibility_timeout,
                    "polling_interval": 1,
                }
                return app

            def test_soft_shutdown_reset_visibility_timeout(self, celery_setup: CeleryTestSetup):
                if isinstance(celery_setup.broker, RabbitMQTestBroker):
                    pytest.skip("RabbitMQ does not support visibility timeout")

                app = celery_setup.app
                queue = celery_setup.worker.worker_queue
                worker = celery_setup.worker
                sig = long_running_task.si(15, verbose=True).set(queue=queue)
                res = sig.delay()

                worker.assert_log_exists("Starting long running task")
                self.kill_worker(worker, WorkerKill.Method.SIGQUIT)
                worker.assert_log_exists(
                    f"Initiating Soft Shutdown, terminating in {app.conf.worker_soft_shutdown_timeout} seconds"
                )
                worker.assert_log_exists("worker: Cold shutdown (MainProcess)")
                worker.assert_log_exists("Restoring 1 unacknowledged message(s)")
                assert_container_exited(worker)
                worker.restart()
                assert res.get(RESULT_TIMEOUT)

            def test_soft_shutdown_reset_visibility_timeout_group_one_finish(self, celery_setup: CeleryTestSetup):
                if isinstance(celery_setup.broker, RabbitMQTestBroker):
                    pytest.skip("RabbitMQ does not support visibility timeout")

                app = celery_setup.app
                queue = celery_setup.worker.worker_queue
                worker = celery_setup.worker
                short_task = long_running_task.si(3, verbose=True).set(queue=queue)
                short_task_res = short_task.freeze()
                long_task = long_running_task.si(15, verbose=True).set(queue=queue)
                long_task_res = long_task.freeze()
                sig = group(short_task, long_task)
                sig.delay()

                worker.assert_log_exists(f"long_running_task[{short_task_res.id}] received")
                worker.assert_log_exists(f"long_running_task[{long_task_res.id}] received")
                self.kill_worker(worker, WorkerKill.Method.SIGQUIT)
                worker.assert_log_exists(
                    f"Initiating Soft Shutdown, terminating in {app.conf.worker_soft_shutdown_timeout} seconds"
                )
                worker.assert_log_exists("worker: Cold shutdown (MainProcess)")
                worker.assert_log_exists("Restoring 1 unacknowledged message(s)")
                assert_container_exited(worker)
                assert short_task_res.get(RESULT_TIMEOUT)

            def test_soft_shutdown_reset_visibility_timeout_group_none_finish(self, celery_setup: CeleryTestSetup):
                if isinstance(celery_setup.broker, RabbitMQTestBroker):
                    pytest.skip("RabbitMQ does not support visibility timeout")

                app = celery_setup.app
                queue = celery_setup.worker.worker_queue
                worker = celery_setup.worker
                short_task = long_running_task.si(15, verbose=True).set(queue=queue)
                short_task_res = short_task.freeze()
                long_task = long_running_task.si(15, verbose=True).set(queue=queue)
                long_task_res = long_task.freeze()
                sig = group(short_task, long_task)
                res = sig.delay()

                worker.assert_log_exists(f"long_running_task[{short_task_res.id}] received")
                worker.assert_log_exists(f"long_running_task[{long_task_res.id}] received")
                self.kill_worker(worker, WorkerKill.Method.SIGQUIT)
                worker.assert_log_exists(
                    f"Initiating Soft Shutdown, terminating in {app.conf.worker_soft_shutdown_timeout} seconds"
                )
                worker.assert_log_exists("worker: Cold shutdown (MainProcess)")
                worker.assert_log_exists("Restoring 2 unacknowledged message(s)")
                assert_container_exited(worker)
                worker.restart()
                assert res.get(RESULT_TIMEOUT) == [True, True]
                assert short_task_res.get(RESULT_TIMEOUT)
                assert long_task_res.get(RESULT_TIMEOUT)

            class test_REMAP_SIGTERM(SuiteOperations):
                @pytest.fixture
                def default_worker_env(self, default_worker_env: dict) -> dict:
                    default_worker_env.update({"REMAP_SIGTERM": "SIGQUIT"})
                    return default_worker_env

                def test_soft_shutdown_reset_visibility_timeout(self, celery_setup: CeleryTestSetup):
                    if isinstance(celery_setup.broker, RabbitMQTestBroker):
                        pytest.skip("RabbitMQ does not support visibility timeout")

                    app = celery_setup.app
                    queue = celery_setup.worker.worker_queue
                    worker = celery_setup.worker
                    sig = long_running_task.si(15, verbose=True).set(queue=queue)
                    res = sig.delay()

                    worker.assert_log_exists("Starting long running task")
                    self.kill_worker(worker, WorkerKill.Method.SIGTERM)
                    worker.assert_log_exists(
                        f"Initiating Soft Shutdown, terminating in {app.conf.worker_soft_shutdown_timeout} seconds"
                    )
                    worker.assert_log_exists("worker: Cold shutdown (MainProcess)")
                    worker.assert_log_exists("Restoring 1 unacknowledged message(s)")
                    assert_container_exited(worker)
                    worker.restart()
                    assert res.get(RESULT_TIMEOUT)

                def test_soft_shutdown_reset_visibility_timeout_group_one_finish(
                    self,
                    celery_setup: CeleryTestSetup,
                ):
                    if isinstance(celery_setup.broker, RabbitMQTestBroker):
                        pytest.skip("RabbitMQ does not support visibility timeout")

                    app = celery_setup.app
                    queue = celery_setup.worker.worker_queue
                    worker = celery_setup.worker
                    short_task = long_running_task.si(3, verbose=True).set(queue=queue)
                    short_task_res = short_task.freeze()
                    long_task = long_running_task.si(15, verbose=True).set(queue=queue)
                    long_task_res = long_task.freeze()
                    sig = group(short_task, long_task)
                    sig.delay()

                    worker.assert_log_exists(f"long_running_task[{short_task_res.id}] received")
                    worker.assert_log_exists(f"long_running_task[{long_task_res.id}] received")
                    self.kill_worker(worker, WorkerKill.Method.SIGTERM)
                    worker.assert_log_exists(
                        f"Initiating Soft Shutdown, terminating in {app.conf.worker_soft_shutdown_timeout} seconds"
                    )
                    worker.assert_log_exists("worker: Cold shutdown (MainProcess)")
                    worker.assert_log_exists("Restoring 1 unacknowledged message(s)")
                    assert_container_exited(worker)
                    assert short_task_res.get(RESULT_TIMEOUT)

            class test_worker_enable_soft_shutdown_on_idle(SuiteOperations):
                @pytest.fixture
                def default_worker_app(self, default_worker_app: Celery) -> Celery:
                    app = default_worker_app
                    app.conf.worker_enable_soft_shutdown_on_idle = True
                    return app

                def test_soft_shutdown(self, celery_setup: CeleryTestSetup):
                    app = celery_setup.app
                    worker = celery_setup.worker

                    self.kill_worker(worker, WorkerKill.Method.SIGQUIT)
                    worker.assert_log_exists(
                        f"Initiating Soft Shutdown, terminating in {app.conf.worker_soft_shutdown_timeout} seconds",
                    )
                    worker.assert_log_exists("worker: Cold shutdown (MainProcess)")

                    assert_container_exited(worker)

                def test_soft_shutdown_eta(self, celery_setup: CeleryTestSetup):
                    if isinstance(celery_setup.broker, RabbitMQTestBroker):
                        pytest.skip("RabbitMQ does not support visibility timeout")

                    app = celery_setup.app
                    queue = celery_setup.worker.worker_queue
                    worker = celery_setup.worker
                    sig = long_running_task.si(5, verbose=True).set(queue=queue)
                    res = sig.apply_async(countdown=app.conf.worker_soft_shutdown_timeout + 5)

                    worker.assert_log_exists(f"long_running_task[{res.id}] received")
                    self.kill_worker(worker, WorkerKill.Method.SIGQUIT)
                    worker.assert_log_exists(
                        f"Initiating Soft Shutdown, terminating in {app.conf.worker_soft_shutdown_timeout} seconds"
                    )
                    worker.assert_log_exists("worker: Cold shutdown (MainProcess)")
                    worker.assert_log_exists("Restoring 1 unacknowledged message(s)")
                    assert_container_exited(worker)
                    worker.restart()
                    assert res.get(RESULT_TIMEOUT)



"""Tests for the ArangoDb."""
import datetime
from unittest.mock import MagicMock, Mock, patch, sentinel

import pytest

from celery.app import backends
from celery.backends import arangodb as module
from celery.backends.arangodb import ArangoDbBackend
from celery.exceptions import ImproperlyConfigured

try:
    import pyArango
except ImportError:
    pyArango = None

pytest.importorskip('pyArango')


class test_ArangoDbBackend:

    def setup_method(self):
        self.backend = ArangoDbBackend(app=self.app)

    def test_init_no_arangodb(self):
        prev, module.py_arango_connection = module.py_arango_connection, None
        try:
            with pytest.raises(ImproperlyConfigured):
                ArangoDbBackend(app=self.app)
        finally:
            module.py_arango_connection = prev

    def test_init_no_settings(self):
        self.app.conf.arangodb_backend_settings = []
        with pytest.raises(ImproperlyConfigured):
            ArangoDbBackend(app=self.app)

    def test_init_settings_is_None(self):
        self.app.conf.arangodb_backend_settings = None
        ArangoDbBackend(app=self.app)

    def test_init_url(self):
        url = None
        expected_database = "celery"
        expected_collection = "celery"
        backend = ArangoDbBackend(app=self.app, url=url)
        assert backend.database == expected_database
        assert backend.collection == expected_collection

        url = "arangodb://localhost:27017/celery-database/celery-collection"
        expected_database = "celery-database"
        expected_collection = "celery-collection"
        backend = ArangoDbBackend(app=self.app, url=url)
        assert backend.database == expected_database
        assert backend.collection == expected_collection

    def test_get_connection_connection_exists(self):
        with patch('pyArango.connection.Connection') as mock_Connection:
            self.backend._connection = sentinel.connection
            connection = self.backend.connection
            assert connection == sentinel.connection
            mock_Connection.assert_not_called()

            expected_connection = mock_Connection()
            mock_Connection.reset_mock()  # So the assert_called_once below is accurate.
            self.backend._connection = None
            connection = self.backend.connection
            assert connection == expected_connection
            mock_Connection.assert_called_once()

    def test_get(self):
        self.backend._connection = MagicMock(spec=["__getitem__"])

        assert self.backend.get(None) is None
        self.backend.db.AQLQuery.assert_not_called()

        assert self.backend.get(sentinel.task_id) is None
        self.backend.db.AQLQuery.assert_called_once_with(
            "RETURN DOCUMENT(@@collection, @key).task",
            rawResults=True,
            bindVars={
                "@collection": self.backend.collection,
                "key": sentinel.task_id,
            },
        )

        self.backend.get = Mock(return_value=sentinel.retval)
        assert self.backend.get(sentinel.task_id) == sentinel.retval
        self.backend.get.assert_called_once_with(sentinel.task_id)

    def test_set(self):
        self.backend._connection = MagicMock(spec=["__getitem__"])

        assert self.backend.set(sentinel.key, sentinel.value) is None
        self.backend.db.AQLQuery.assert_called_once_with(
            """
            UPSERT {_key: @key}
            INSERT {_key: @key, task: @value}
            UPDATE {task: @value} IN @@collection
            """,
            bindVars={
                "@collection": self.backend.collection,
                "key": sentinel.key,
                "value": sentinel.value,
            },
        )

    def test_mget(self):
        self.backend._connection = MagicMock(spec=["__getitem__"])

        result = list(self.backend.mget(None))
        expected_result = []
        assert result == expected_result
        self.backend.db.AQLQuery.assert_not_called()

        Query = MagicMock(spec=pyArango.query.Query)
        query = Query()
        query.nextBatch = MagicMock(side_effect=StopIteration())
        self.backend.db.AQLQuery = Mock(return_value=query)

        keys = [sentinel.task_id_0, sentinel.task_id_1]
        result = list(self.backend.mget(keys))
        expected_result = []
        assert result == expected_result
        self.backend.db.AQLQuery.assert_called_once_with(
            "FOR k IN @keys RETURN DOCUMENT(@@collection, k).task",
            rawResults=True,
            bindVars={
                "@collection": self.backend.collection,
                "keys": keys,
            },
        )

        values = [sentinel.value_0, sentinel.value_1]
        query.__iter__.return_value = iter([sentinel.value_0, sentinel.value_1])
        result = list(self.backend.mget(keys))
        expected_result = values
        assert result == expected_result

    def test_delete(self):
        self.backend._connection = MagicMock(spec=["__getitem__"])

        assert self.backend.delete(None) is None
        self.backend.db.AQLQuery.assert_not_called()

        assert self.backend.delete(sentinel.task_id) is None
        self.backend.db.AQLQuery.assert_called_once_with(
            "REMOVE {_key: @key} IN @@collection",
            bindVars={
                "@collection": self.backend.collection,
                "key": sentinel.task_id,
            },
        )

    def test_config_params(self):
        self.app.conf.arangodb_backend_settings = {
            'host': 'test.arangodb.com',
            'port': '8529',
            'username': 'johndoe',
            'password': 'mysecret',
            'database': 'celery_database',
            'collection': 'celery_collection',
            'http_protocol': 'https',
            'verify': True
        }
        x = ArangoDbBackend(app=self.app)
        assert x.host == 'test.arangodb.com'
        assert x.port == 8529
        assert x.username == 'johndoe'
        assert x.password == 'mysecret'
        assert x.database == 'celery_database'
        assert x.collection == 'celery_collection'
        assert x.http_protocol == 'https'
        assert x.arangodb_url == 'https://test.arangodb.com:8529'
        assert x.verify is True

    def test_backend_by_url(
        self, url="arangodb://username:password@host:port/database/collection"
    ):
        from celery.backends.arangodb import ArangoDbBackend
        backend, url_ = backends.by_url(url, self.app.loader)
        assert backend is ArangoDbBackend
        assert url_ == url

    def test_backend_params_by_url(self):
        url = (
            "arangodb://johndoe:mysecret@test.arangodb.com:8529/"
            "celery_database/celery_collection"
        )
        with self.Celery(backend=url) as app:
            x = app.backend
            assert x.host == 'test.arangodb.com'
            assert x.port == 8529
            assert x.username == 'johndoe'
            assert x.password == 'mysecret'
            assert x.database == 'celery_database'
            assert x.collection == 'celery_collection'
            assert x.http_protocol == 'http'
            assert x.arangodb_url == 'http://test.arangodb.com:8529'
            assert x.verify is False

    def test_backend_cleanup(self):
        self.backend._connection = MagicMock(spec=["__getitem__"])

        self.backend.expires = None
        self.backend.cleanup()
        self.backend.db.AQLQuery.assert_not_called()

        self.backend.expires = 0
        self.backend.cleanup()
        self.backend.db.AQLQuery.assert_not_called()

        now = datetime.datetime.now(datetime.timezone.utc)
        self.backend.app.now = Mock(return_value=now)
        self.backend.expires = 86400
        expected_checkpoint = (now - self.backend.expires_delta).isoformat()
        self.backend.cleanup()
        self.backend.db.AQLQuery.assert_called_once_with(
            """
            FOR record IN @@collection
                FILTER record.task.date_done < @checkpoint
                REMOVE record IN @@collection
            """,
            bindVars={
                "@collection": self.backend.collection,
                "checkpoint": expected_checkpoint,
            },
        )



import os
import re
from unittest.mock import patch

import pytest
from click.testing import CliRunner
from kombu.exceptions import OperationalError

from celery.bin.celery import celery
from celery.platforms import EX_UNAVAILABLE

_GLOBAL_OPTIONS = ['-A', 't.unit.bin.proj.app_with_custom_cmds', '--broker', 'memory://']
_INSPECT_OPTIONS = ['--timeout', '0']  # Avoid waiting for the zero workers to reply


@pytest.fixture(autouse=True)
def clean_os_environ():
    # Celery modifies os.environ when given the CLI option --broker memory://
    # This interferes with other tests, so we need to reset os.environ
    with patch.dict(os.environ, clear=True):
        yield


@pytest.mark.parametrize(
    ('celery_cmd', 'custom_cmd'),
    [
        ('inspect', ('custom_inspect_cmd', '123')),
        ('control', ('custom_control_cmd', '123', '456')),
    ],
)
def test_custom_remote_command(celery_cmd, custom_cmd, isolated_cli_runner: CliRunner):
    res = isolated_cli_runner.invoke(
        celery,
        [*_GLOBAL_OPTIONS, celery_cmd, *_INSPECT_OPTIONS, *custom_cmd],
        catch_exceptions=False,
    )
    assert res.exit_code == EX_UNAVAILABLE, (res, res.output)
    assert res.output.strip() == 'Error: No nodes replied within time constraint'


@pytest.mark.parametrize(
    ('celery_cmd', 'remote_cmd'),
    [
        # Test nonexistent commands
        ('inspect', 'this_command_does_not_exist'),
        ('control', 'this_command_does_not_exist'),
        # Test commands that exist, but are of the wrong type
        ('inspect', 'custom_control_cmd'),
        ('control', 'custom_inspect_cmd'),
    ],
)
def test_unrecognized_remote_command(celery_cmd, remote_cmd, isolated_cli_runner: CliRunner):
    res = isolated_cli_runner.invoke(
        celery,
        [*_GLOBAL_OPTIONS, celery_cmd, *_INSPECT_OPTIONS, remote_cmd],
        catch_exceptions=False,
    )
    assert res.exit_code == 2, (res, res.output)
    assert f'Error: Command {remote_cmd} not recognized. Available {celery_cmd} commands: ' in res.output


_expected_inspect_regex = (
    '\n  custom_inspect_cmd x\\s+Ask the workers to reply with x\\.\n'
)
_expected_control_regex = (
    '\n  custom_control_cmd a b\\s+Ask the workers to reply with a and b\\.\n'
)


@pytest.mark.parametrize(
    ('celery_cmd', 'expected_regex'),
    [
        ('inspect', re.compile(_expected_inspect_regex, re.MULTILINE)),
        ('control', re.compile(_expected_control_regex, re.MULTILINE)),
    ],
)
def test_listing_remote_commands(celery_cmd, expected_regex, isolated_cli_runner: CliRunner):
    res = isolated_cli_runner.invoke(
        celery,
        [*_GLOBAL_OPTIONS, celery_cmd, '--list'],
    )
    assert res.exit_code == 0, (res, res.stdout)
    assert expected_regex.search(res.stdout)


def test_status_shows_friendly_error_when_broker_unreachable(cli_runner: CliRunner):
    with patch('celery.app.control.Inspect.ping',
               side_effect=OperationalError('[Errno 61] Connection refused')):
        res = cli_runner.invoke(
            celery,
            [*_GLOBAL_OPTIONS, 'status'],
            catch_exceptions=False,
        )
    assert res.exit_code == EX_UNAVAILABLE, (res, res.output)
    assert 'Error: Could not connect to the message broker.' in res.output
    assert 'Reason: [Errno 61] Connection refused' in res.output
    assert 'Traceback' not in res.output


def test_status_unexpected_error_is_summarized(cli_runner: CliRunner):
    with patch('celery.app.control.Inspect.ping',
               side_effect=RuntimeError('boom')):
        res = cli_runner.invoke(
            celery,
            [*_GLOBAL_OPTIONS, 'status'],
            catch_exceptions=False,
        )
    assert res.exit_code == EX_UNAVAILABLE, (res, res.output)
    assert 'Error: Unable to run the `status` command. Reason: boom' in res.output
    assert 'Traceback' not in res.output


def test_graph_workers_shows_friendly_error_when_broker_unreachable(
    cli_runner: CliRunner,
):
    with patch('celery.app.control.Inspect.stats',
               side_effect=OperationalError('connection failed')):
        res = cli_runner.invoke(
            celery,
            [*_GLOBAL_OPTIONS, 'graph', 'workers'],
            catch_exceptions=False,
        )
    assert res.exit_code == EX_UNAVAILABLE, (res, res.output)
    assert 'Error: Could not connect to the message broker.' in res.output
    assert 'Reason: connection failed' in res.output


def test_events_dump_shows_friendly_error_when_broker_unreachable(
    cli_runner: CliRunner,
):
    with patch('celery.bin.events._run_evdump',
               side_effect=OperationalError('connection failed')):
        res = cli_runner.invoke(
            celery,
            [*_GLOBAL_OPTIONS, 'events', '--dump'],
            catch_exceptions=False,
        )
    assert res.exit_code == EX_UNAVAILABLE, (res, res.output)
    assert 'Error: Could not connect to the message broker.' in res.output
    assert 'Reason: connection failed' in res.output


def test_handle_remote_command_error_reraises_click_exception():
    """base.py: bare ``raise`` inside the ClickException branch must be covered."""
    import click

    from celery.bin.base import handle_remote_command_error

    original = click.ClickException('original click error')
    with pytest.raises(click.ClickException) as exc_info:
        try:
            raise original
        except Exception as exc:
            handle_remote_command_error('any', exc)
    assert exc_info.value is original


def test_inspect_shows_friendly_error_when_broker_unreachable(cli_runner: CliRunner):
    with patch('celery.app.control.Inspect._request',
               side_effect=OperationalError('connection refused')):
        res = cli_runner.invoke(
            celery,
            [*_GLOBAL_OPTIONS, 'inspect', *_INSPECT_OPTIONS, 'custom_inspect_cmd', '1'],
            catch_exceptions=False,
        )
    assert res.exit_code == EX_UNAVAILABLE, (res, res.output)
    assert 'Error: Could not connect to the message broker.' in res.output
    assert 'Reason: connection refused' in res.output
    assert 'Traceback' not in res.output


def test_inspect_unexpected_error_is_summarized(cli_runner: CliRunner):
    with patch('celery.app.control.Inspect._request',
               side_effect=RuntimeError('inspect boom')):
        res = cli_runner.invoke(
            celery,
            [*_GLOBAL_OPTIONS, 'inspect', *_INSPECT_OPTIONS, 'custom_inspect_cmd', '1'],
            catch_exceptions=False,
        )
    assert res.exit_code == EX_UNAVAILABLE, (res, res.output)
    assert 'Error: Unable to run the `inspect custom_inspect_cmd` command. Reason: inspect boom' in res.output
    assert 'Traceback' not in res.output


def test_control_shows_friendly_error_when_broker_unreachable(cli_runner: CliRunner):
    with patch('celery.app.control.Control.broadcast',
               side_effect=OperationalError('connection refused')):
        res = cli_runner.invoke(
            celery,
            [*_GLOBAL_OPTIONS, 'control', *_INSPECT_OPTIONS, 'custom_control_cmd', '1', '2'],
            catch_exceptions=False,
        )
    assert res.exit_code == EX_UNAVAILABLE, (res, res.output)
    assert 'Error: Could not connect to the message broker.' in res.output
    assert 'Reason: connection refused' in res.output
    assert 'Traceback' not in res.output


def test_control_unexpected_error_is_summarized(cli_runner: CliRunner):
    with patch('celery.app.control.Control.broadcast',
               side_effect=RuntimeError('control boom')):
        res = cli_runner.invoke(
            celery,
            [*_GLOBAL_OPTIONS, 'control', *_INSPECT_OPTIONS, 'custom_control_cmd', '1', '2'],
            catch_exceptions=False,
        )
    assert res.exit_code == EX_UNAVAILABLE, (res, res.output)
    assert 'Error: Unable to run the `control custom_control_cmd` command. Reason: control boom' in res.output
    assert 'Traceback' not in res.output


def test_events_camera_shows_friendly_error_when_broker_unreachable(cli_runner: CliRunner):
    with patch('celery.bin.events._run_evcam',
               side_effect=OperationalError('connection failed')):
        res = cli_runner.invoke(
            celery,
            [*_GLOBAL_OPTIONS, 'events', '--camera', 'myapp.MyCameraClass'],
            catch_exceptions=False,
        )
    assert res.exit_code == EX_UNAVAILABLE, (res, res.output)
    assert 'Error: Could not connect to the message broker.' in res.output
    assert 'Reason: connection failed' in res.output
    assert 'Traceback' not in res.output


def test_events_evtop_shows_friendly_error_when_broker_unreachable(cli_runner: CliRunner):
    with patch('celery.bin.events._run_evtop',
               side_effect=OperationalError('connection failed')):
        res = cli_runner.invoke(
            celery,
            [*_GLOBAL_OPTIONS, 'events'],
            catch_exceptions=False,
        )
    assert res.exit_code == EX_UNAVAILABLE, (res, res.output)
    assert 'Error: Could not connect to the message broker.' in res.output
    assert 'Reason: connection failed' in res.output
    assert 'Traceback' not in res.output



from unittest.mock import MagicMock, Mock

from click import Option

from celery import Celery

# This module defines a mocked Celery application to replicate
# the behavior of Pyramid-Celery's configuration by preload options.
# Preload options should propagate to commands like shell and purge etc.
#
# The Pyramid-Celery project https://github.com/sontek/pyramid_celery
# assumes that you want to configure Celery via an ini settings file.
# The .ini files are the standard configuration file for Pyramid
# applications.
# See https://docs.pylonsproject.org/projects/pyramid/en/latest/quick_tutorial/ini.html
#

app = Celery(set_as_current=False)
app.config_from_object("t.integration.test_worker_config")


class PurgeMock:
    def queue_purge(self, queue):
        return 0


class ConnMock:
    default_channel = PurgeMock()
    channel_errors = KeyError


mock = Mock()
mock.__enter__ = Mock(return_value=ConnMock())
mock.__exit__ = Mock(return_value=False)

app.connection_for_write = MagicMock(return_value=mock)

# Below are taken from pyramid-celery's __init__.py
# Ref: https://github.com/sontek/pyramid_celery/blob/cf8aa80980e42f7235ad361874d3c35e19963b60/pyramid_celery/__init__.py#L25-L36 # noqa: E501
ini_option = Option(
    (
        "--ini",
        "-i",
    ),
    help="Paste ini configuration file.",
)

ini_var_option = Option(
    ("--ini-var",), help="Comma separated list of key=value to pass to ini."
)

app.user_options["preload"].add(ini_option)
app.user_options["preload"].add(ini_var_option)



import pytest
from pytest_celery import CeleryBackendCluster, CeleryTestSetup

from celery.signals import after_task_publish, before_task_publish
from t.smoke.tasks import noop


@pytest.fixture
def default_worker_signals(default_worker_signals: set) -> set:
    from t.smoke import signals

    default_worker_signals.add(signals)
    yield default_worker_signals


@pytest.fixture
def celery_backend_cluster() -> CeleryBackendCluster:
    # Disable backend
    return None


class test_signals:
    @pytest.mark.parametrize(
        "log, control",
        [
            ("worker_init_handler", None),
            ("worker_process_init_handler", None),
            ("worker_ready_handler", None),
            ("worker_process_shutdown_handler", "shutdown"),
            ("worker_shutdown_handler", "shutdown"),
        ],
    )
    def test_sanity(self, celery_setup: CeleryTestSetup, log: str, control: str):
        if control:
            celery_setup.app.control.broadcast(control)
        celery_setup.worker.wait_for_log(log)


class test_before_task_publish:
    def test_sanity(self, celery_setup: CeleryTestSetup):
        @before_task_publish.connect
        def before_task_publish_handler(*args, **kwargs):
            nonlocal signal_was_called
            signal_was_called = True

        signal_was_called = False
        noop.s().apply_async(queue=celery_setup.worker.worker_queue)
        assert signal_was_called is True


class test_after_task_publish:
    def test_sanity(self, celery_setup: CeleryTestSetup):
        @after_task_publish.connect
        def after_task_publish_handler(*args, **kwargs):
            nonlocal signal_was_called
            signal_was_called = True

        signal_was_called = False
        noop.s().apply_async(queue=celery_setup.worker.worker_queue)
        assert signal_was_called is True



import json
import logging
import os
import re
import time

import pytest

from celery.contrib.pytest import celery_app, celery_session_worker
from celery.contrib.testing.manager import Manager
from celery.exceptions import TimeoutError
from t.integration.tasks import get_redis_connection

# we have to import the pytest plugin fixtures here,
# in case user did not do the `python setup.py develop` yet,
# that installs the pytest plugin into the setuptools registry.


logger = logging.getLogger(__name__)

TEST_BROKER = os.environ.get('TEST_BROKER', 'pyamqp://')
TEST_BACKEND = os.environ.get('TEST_BACKEND', 'redis://')

RETRYABLE_EXCEPTIONS = (OSError, ConnectionError, TimeoutError)


def is_retryable_exception(exc):
    return isinstance(exc, RETRYABLE_EXCEPTIONS)


_flaky = pytest.mark.flaky(reruns=5, reruns_delay=1, cause=is_retryable_exception)
_timeout = pytest.mark.timeout(timeout=300)


def flaky(fn):
    return _timeout(_flaky(fn))


__all__ = (
    'celery_app',
    'celery_session_worker',
    'flaky',
    'get_active_redis_channels',
)


def get_active_redis_channels():
    return get_redis_connection().execute_command('PUBSUB CHANNELS')


def check_for_logs(caplog, message: str, max_wait: float = 1.0, interval: float = 0.1) -> bool:
    start_time = time.monotonic()
    while time.monotonic() - start_time < max_wait:
        if any(re.search(message, record.message) for record in caplog.records):
            return True
        time.sleep(interval)
    return False


@pytest.fixture(scope='session')
def celery_config(request):
    config = {
        'broker_url': TEST_BROKER,
        'result_backend': TEST_BACKEND,
        'result_extended': True,
        'cassandra_servers': ['localhost'],
        'cassandra_keyspace': 'tests',
        'cassandra_table': 'tests',
        'cassandra_read_consistency': 'ONE',
        'cassandra_write_consistency': 'ONE',
    }
    try:
        # To override the default configuration, create the integration-tests-config.json file
        # in Celery's root directory.
        # The file must contain a dictionary of valid configuration name/value pairs.
        with open(str(request.config.rootdir / "integration-tests-config.json")) as file:
            overrides = json.load(file)
        config.update(overrides)
    except OSError:
        pass
    return config


@pytest.fixture(scope='session')
def celery_enable_logging():
    return True


@pytest.fixture(scope='session')
def celery_worker_pool():
    return 'prefork'


@pytest.fixture(scope='session')
def celery_includes():
    return {'t.integration.tasks'}


@pytest.fixture
def app(celery_app):
    yield celery_app


@pytest.fixture
def manager(app, celery_session_worker):
    manager = Manager(app)
    yield manager
    try:
        manager.wait_until_idle()
    except Exception as e:
        logger.warning("Failed to stop Celery test manager cleanly: %s", e)


@pytest.fixture(autouse=True)
def ZZZZ_set_app_current(app):
    app.set_current()
    app.set_default()


@pytest.fixture(scope='session')
def celery_class_tasks():
    from t.integration.tasks import ClassBasedAutoRetryTask
    return [ClassBasedAutoRetryTask]



import sys
from time import monotonic
from unittest.mock import Mock, patch

import pytest

from celery.concurrency.base import BasePool
from celery.utils.objects import Bunch
from celery.worker import autoscale, state


class MockPool(BasePool):

    shrink_raises_exception = False
    shrink_raises_ValueError = False

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._pool = Bunch(_processes=self.limit)

    def grow(self, n=1):
        self._pool._processes += n

    def shrink(self, n=1):
        if self.shrink_raises_exception:
            raise KeyError('foo')
        if self.shrink_raises_ValueError:
            raise ValueError('foo')
        self._pool._processes -= n

    @property
    def num_processes(self):
        return self._pool._processes


class test_WorkerComponent:

    def test_register_with_event_loop(self):
        parent = Mock(name='parent')
        parent.autoscale = True
        parent.consumer.on_task_message = set()
        w = autoscale.WorkerComponent(parent)
        assert parent.autoscaler is None
        assert w.enabled

        hub = Mock(name='hub')
        w.create(parent)
        w.register_with_event_loop(parent, hub)
        assert (parent.autoscaler.maybe_scale in
                parent.consumer.on_task_message)
        hub.call_repeatedly.assert_called_with(
            parent.autoscaler.keepalive, parent.autoscaler.maybe_scale,
        )

        parent.hub = hub
        hub.on_init = []
        w.instantiate = Mock()
        w.register_with_event_loop(parent, Mock(name='loop'))
        assert parent.consumer.on_task_message

    def test_info_without_event_loop(self):
        parent = Mock(name='parent')
        parent.autoscale = True
        parent.max_concurrency = '10'
        parent.min_concurrency = '2'
        parent.use_eventloop = False
        w = autoscale.WorkerComponent(parent)
        w.create(parent)
        info = w.info(parent)
        assert 'autoscaler' in info
        assert parent.autoscaler_cls().info.called


class test_Autoscaler:

    def setup_method(self):
        self.pool = MockPool(3)

    def test_stop(self):

        class Scaler(autoscale.Autoscaler):
            alive = True
            joined = False

            def is_alive(self):
                return self.alive

            def join(self, timeout=None):
                self.joined = True

        worker = Mock(name='worker')
        x = Scaler(self.pool, 10, 3, worker=worker)
        # Don't allow thread joining or event waiting to block the test
        with patch("threading.Thread.join"), patch("threading.Event.wait"):
            x.stop()
        assert x.joined
        x.joined = False
        x.alive = False
        with patch("threading.Thread.join"), patch("threading.Event.wait"):
            x.stop()
        assert not x.joined

    @pytest.mark.sleepdeprived_patched_module(autoscale)
    def test_body(self, sleepdeprived):
        worker = Mock(name='worker')
        x = autoscale.Autoscaler(self.pool, 10, 3, worker=worker)
        x.body()
        assert x.pool.num_processes == 3
        _keep = [Mock(name=f'req{i}') for i in range(20)]
        [state.task_reserved(m) for m in _keep]
        x.body()
        x.body()
        assert x.pool.num_processes == 10
        state.reserved_requests.clear()
        x.body()
        assert x.pool.num_processes == 10
        x._last_scale_up = monotonic() - 10000
        x.body()
        assert x.pool.num_processes == 3

    def test_run(self):

        class Scaler(autoscale.Autoscaler):
            scale_called = False

            def body(self):
                self.scale_called = True
                getattr(self, "_bgThread__is_shutdown").set()

        worker = Mock(name='worker')
        x = Scaler(self.pool, 10, 3, worker=worker)
        x.run()
        assert getattr(x, "_bgThread__is_shutdown").is_set()
        assert getattr(x, "_bgThread__is_stopped").is_set()
        assert x.scale_called

    def test_shrink_raises_exception(self):
        worker = Mock(name='worker')
        x = autoscale.Autoscaler(self.pool, 10, 3, worker=worker)
        x.scale_up(3)
        x.pool.shrink_raises_exception = True
        x._shrink(1)

    @patch('celery.worker.autoscale.debug')
    def test_shrink_raises_ValueError(self, debug):
        worker = Mock(name='worker')
        x = autoscale.Autoscaler(self.pool, 10, 3, worker=worker)
        x.scale_up(3)
        x._last_scale_up = monotonic() - 10000
        x.pool.shrink_raises_ValueError = True
        x.scale_down(1)
        assert debug.call_count

    def test_update(self):
        worker = Mock(name='worker')
        x = autoscale.Autoscaler(self.pool, 10, 3, worker=worker)
        x.worker.consumer.prefetch_multiplier = 1
        x.keepalive = -1
        assert x.processes == 3
        x.scale_up(5)
        x.update(7, None)
        assert x.processes == 7
        assert x.max_concurrency == 7
        x.scale_down(4)
        x.update(None, 6)
        assert x.processes == 6
        assert x.min_concurrency == 6

        x.update(max=300, min=10)
        x.update(max=300, min=2)
        x.update(max=None, min=None)

    def test_prefetch_count_on_updates(self):
        worker = Mock(name='worker')
        x = autoscale.Autoscaler(self.pool, 10, 3, worker=worker)
        x.worker.consumer.prefetch_multiplier = 1
        x.update(5, None)
        worker.consumer._update_prefetch_count.assert_called_with(-5)
        x.update(15, 7)
        worker.consumer._update_prefetch_count.assert_called_with(10)

    def test_prefetch_count_on_updates_prefetch_multiplier_gt_one(self):
        worker = Mock(name='worker')
        x = autoscale.Autoscaler(self.pool, 10, 3, worker=worker)
        x.worker.consumer.prefetch_multiplier = 4
        x.update(5, None)
        worker.consumer._update_prefetch_count.assert_called_with(-5)
        x.update(15, 7)
        worker.consumer._update_prefetch_count.assert_called_with(10)

    def test_info(self):
        worker = Mock(name='worker')
        x = autoscale.Autoscaler(self.pool, 10, 3, worker=worker)
        info = x.info()
        assert info['max'] == 10
        assert info['min'] == 3
        assert info['current'] == 3

    @patch('os._exit')
    def test_thread_crash(self, _exit):

        class _Autoscaler(autoscale.Autoscaler):

            def body(self):
                getattr(self, "_bgThread__is_shutdown").set()
                raise OSError('foo')
        worker = Mock(name='worker')
        x = _Autoscaler(self.pool, 10, 3, worker=worker)

        stderr = Mock()
        p, sys.stderr = sys.stderr, stderr
        try:
            x.run()
        finally:
            sys.stderr = p
        _exit.assert_called_with(1)
        stderr.write.assert_called()

    @pytest.mark.sleepdeprived_patched_module(autoscale)
    def test_no_negative_scale(self, sleepdeprived):
        total_num_processes = []
        worker = Mock(name='worker')
        x = autoscale.Autoscaler(self.pool, 10, 3, worker=worker)
        x.body()  # the body func scales up or down

        _keep = [Mock(name=f'req{i}') for i in range(35)]
        for req in _keep:
            state.task_reserved(req)
            x.body()
            total_num_processes.append(self.pool.num_processes)

        for req in _keep:
            state.task_ready(req)
            x.body()
            total_num_processes.append(self.pool.num_processes)

        assert all(x.min_concurrency <= i <= x.max_concurrency
                   for i in total_num_processes)

    def test_disable_prefetch_respects_max_concurrency(self):
        """Test that disable_prefetch respects autoscale max_concurrency setting"""
        from celery.worker.consumer.tasks import Tasks

        # Create a mock consumer with autoscale and disable_prefetch enabled
        consumer = Mock()
        consumer.app = Mock()
        consumer.app.conf.worker_disable_prefetch = True
        consumer.pool = Mock()
        consumer.pool.num_processes = 10
        consumer.controller = Mock()
        consumer.controller.max_concurrency = 5  # Lower than pool processes

        # Mock task consumer setup
        consumer.task_consumer = Mock()
        consumer.task_consumer.channel = Mock()
        consumer.task_consumer.channel.qos = Mock()
        consumer.task_consumer.channel.qos.can_consume = Mock(return_value=True)

        # Mock the connection and other required attributes
        consumer.connection = Mock()
        consumer.connection.default_channel = Mock()
        consumer.connection.transport = Mock()
        consumer.connection.transport.driver_type = 'redis'
        consumer.initial_prefetch_count = 20
        consumer.update_strategies = Mock()
        consumer.on_decode_error = Mock()

        # Mock the amqp TaskConsumer
        consumer.app.amqp = Mock()
        consumer.app.amqp.TaskConsumer = Mock(return_value=consumer.task_consumer)

        tasks_instance = Tasks(consumer)

        # Mock 5 reserved requests (at autoscale limit of 5)
        mock_requests = [Mock() for _ in range(5)]
        with patch('celery.worker.state.reserved_requests', mock_requests):
            tasks_instance.start(consumer)

            # Should not be able to consume when at autoscale limit
            assert consumer.task_consumer.channel.qos.can_consume() is False

        # Test with 4 reserved requests (under autoscale limit of 5)
        mock_requests = [Mock() for _ in range(4)]
        with patch('celery.worker.state.reserved_requests', mock_requests):
            tasks_instance.start(consumer)

            # Should be able to consume when under autoscale limit
            assert consumer.task_consumer.channel.qos.can_consume() is True



import requests
from pytest_celery import RESULT_TIMEOUT, CeleryTestSetup
from requests.auth import HTTPBasicAuth

from celery.canvas import group
from t.integration.tasks import add, identity
from t.smoke.tests.quorum_queues.conftest import RabbitMQManagementBroker


class test_broker_configuration:
    def test_queue_type(self, celery_setup: CeleryTestSetup):
        broker: RabbitMQManagementBroker = celery_setup.broker
        api = broker.get_management_url() + "/api/queues"
        response = requests.get(api, auth=HTTPBasicAuth("guest", "guest"))
        assert response.status_code == 200
        res = response.json()
        assert isinstance(res, list)
        worker_queue = next((queue for queue in res if queue["name"] == celery_setup.worker.worker_queue), None)
        assert worker_queue is not None, f'"{celery_setup.worker.worker_queue}" queue not found'
        queue_type = worker_queue.get("type")
        assert queue_type == "quorum", f'"{celery_setup.worker.worker_queue}" queue is not a quorum queue'


class test_quorum_queues:
    def test_signature(self, celery_setup: CeleryTestSetup):
        sig = identity.si("test_signature").set(queue=celery_setup.worker.worker_queue)
        assert sig.delay().get(timeout=RESULT_TIMEOUT) == "test_signature"

    def test_group(self, celery_setup: CeleryTestSetup):
        sig = group(
            group(add.si(1, 1), add.si(2, 2)),
            group([add.si(1, 1), add.si(2, 2)]),
            group(s for s in [add.si(1, 1), add.si(2, 2)]),
        )
        res = sig.apply_async(queue=celery_setup.worker.worker_queue)
        assert res.get(timeout=RESULT_TIMEOUT) == [2, 4, 2, 4, 2, 4]



from __future__ import annotations

import json

from pytest_celery import RESULT_TIMEOUT, CeleryTestWorker

from celery.canvas import StampingVisitor
from t.integration.tasks import add, identity


class test_stamping_visitor:
    def test_callback(self, dev_worker: CeleryTestWorker):
        on_signature_stamp = {"on_signature_stamp": 4}
        no_visitor_stamp = {"no_visitor_stamp": "Stamp without visitor"}
        on_callback_stamp = {"on_callback_stamp": 2}
        link_stamp = {
            **on_signature_stamp,
            **no_visitor_stamp,
            **on_callback_stamp,
        }

        class CustomStampingVisitor(StampingVisitor):
            def on_signature(self, sig, **headers) -> dict:
                return on_signature_stamp.copy()

            def on_callback(self, callback, **header) -> dict:
                return on_callback_stamp.copy()

        stamped_task = identity.si(123).set(queue=dev_worker.worker_queue)
        stamped_task.link(
            add.s(0)
            .stamp(no_visitor_stamp=no_visitor_stamp["no_visitor_stamp"])
            .set(queue=dev_worker.worker_queue)
        )
        stamped_task.stamp(visitor=CustomStampingVisitor())
        stamped_task.delay().get(timeout=RESULT_TIMEOUT)
        assert dev_worker.logs().count(
            json.dumps(on_signature_stamp, indent=4, sort_keys=True)
        )
        assert dev_worker.logs().count(json.dumps(link_stamp, indent=4, sort_keys=True))



from time import sleep

from config import app
from visitors import FullVisitor, MonitoringIdStampingVisitor, MyStampingVisitor

from celery import Task
from celery.canvas import Signature, maybe_signature
from celery.utils.log import get_task_logger

logger = get_task_logger(__name__)


def log_demo(running_task):
    request, name = running_task.request, running_task.name + running_task.request.argsrepr
    if hasattr(request, "stamps"):
        stamps = request.stamps or {}
        stamped_headers = request.stamped_headers or []

    if stamps and stamped_headers:
        logger.critical(f"Found {name}.stamps: {stamps}")
        logger.critical(f"Found {name}.stamped_headers: {stamped_headers}")
    else:
        logger.critical(f"Running {name} without stamps")

    links = request.callbacks or []
    for link in links:
        link = maybe_signature(link)
        logger.critical(f"Found {name}.link: {link}")
        stamped_headers = link.options.get("stamped_headers", [])
        stamps = {stamp: link.options[stamp] for stamp in stamped_headers}

        if stamps and stamped_headers:
            logger.critical(f"Found {name}.link stamps: {stamps}")
            logger.critical(f"Found {name}.link stamped_headers: {stamped_headers}")
        else:
            logger.critical(f"Running {name}.link without stamps")


class StampOnReplace(Task):
    """Custom task for stamping on replace"""

    def on_replace(self, sig: Signature):
        logger.warning(f"StampOnReplace: {sig}.stamp(FullVisitor())")
        sig.stamp(FullVisitor())
        logger.warning(f"StampOnReplace: {sig}.stamp(MyStampingVisitor())")
        sig.stamp(MyStampingVisitor())
        return super().on_replace(sig)


class MonitoredTask(Task):
    def on_replace(self, sig: Signature):
        logger.warning(f"MonitoredTask: {sig}.stamp(MonitoringIdStampingVisitor())")
        sig.stamp(MonitoringIdStampingVisitor(), append_stamps=False)
        return super().on_replace(sig)


@app.task(bind=True)
def identity_task(self, x):
    """Identity function"""
    log_demo(self)
    return x


@app.task(bind=True, base=MonitoredTask)
def replaced_identity(self: MonitoredTask, x):
    log_demo(self)
    logger.warning("Stamping identity_task with MonitoringIdStampingVisitor() before replace")
    replaced_task = identity_task.s(x)
    # These stamps should be overridden by the stamps from MonitoredTask.on_replace()
    replaced_task.stamp(MonitoringIdStampingVisitor())
    return self.replace(replaced_task)


@app.task(bind=True, base=StampOnReplace)
def identity(self: Task, x):
    log_demo(self)
    return self.replace(replaced_identity.s(x))


@app.task
def mul(x: int, y: int) -> int:
    """Multiply two numbers"""
    return x * y


@app.task
def xsum(numbers: list) -> int:
    """Sum a list of numbers"""
    return sum(numbers)


@app.task
def waitfor(seconds: int) -> None:
    """Wait for "seconds" seconds, ticking every second."""
    print(f"Waiting for {seconds} seconds...")
    for i in range(seconds):
        sleep(1)
        print(f"{i+1} seconds passed")


@app.task(bind=True, base=StampOnReplace)
def wait_for_revoke(self: StampOnReplace, seconds: int) -> None:
    """Replace this task with a new task that waits for "seconds" seconds."""
    self.replace(waitfor.s(seconds))



#!/usr/bin/env python3

import re
import sys

import click
import pyperclip
from colorama import Fore, init

# Initialize colorama for color support in terminal
init(autoreset=True)

# Regular expression pattern to match the required lines
PATTERN = re.compile(r"^\*\s*(.*?)\s+by\s+@[\w-]+\s+in\s+https://github\.com/[\w-]+/[\w-]+/pull/(\d+)")


def read_changes_file(filename):
    try:
        with open(filename) as f:
            return f.readlines()
    except FileNotFoundError:
        print(f"Error: {filename} file not found.")
        sys.exit(1)


def read_from_clipboard():
    text = pyperclip.paste()
    return text.splitlines()


def process_line(line):
    line = line.strip()

    # Skip lines containing '[pre-commit.ci]'
    if "[pre-commit.ci]" in line:
        return None

    # Skip lines starting with '## What's Changed'
    if line.startswith("## What's Changed"):
        return None

    # Stop processing if '## New Contributors' is encountered
    if line.startswith("## New Contributors"):
        return "STOP_PROCESSING"

    # Skip lines that don't start with '* '
    if not line.startswith("* "):
        return None

    match = PATTERN.match(line)
    if match:
        description, pr_number = match.groups()
        return f"- {description} (#{pr_number})"
    return None


@click.command()
@click.option(
    "--source",
    "-s",
    type=click.Path(exists=True),
    help="Source file to read from. If not provided, reads from clipboard.",
)
@click.option(
    "--dest",
    "-d",
    type=click.File("w"),
    default="-",
    help="Destination file to write to. Defaults to standard output.",
)
@click.option(
    "--clipboard",
    "-c",
    is_flag=True,
    help="Read input from clipboard explicitly.",
)
def main(source, dest, clipboard):
    # Determine the source of input
    if clipboard or (not source and not sys.stdin.isatty()):
        # Read from clipboard
        lines = read_from_clipboard()
    elif source:
        # Read from specified file
        lines = read_changes_file(source)
    else:
        # Default: read from clipboard
        lines = read_from_clipboard()

    output_lines = []
    for line in lines:
        output_line = process_line(line)
        if output_line == "STOP_PROCESSING":
            break
        if output_line:
            output_lines.append(output_line)

    output_text = "\n".join(output_lines)

    # Prepare the header
    version = "x.y.z"
    underline = "=" * len(version)

    header = f"""
.. _version-{version}:

{version}
{underline}

:release-date: <YYYY-MM-DD>
:release-by: <FULL NAME>

What's Changed
~~~~~~~~~~~~~~
"""

    # Combine header and output
    final_output = header + output_text

    # Write output to destination
    if dest.name == "<stdout>":
        print(Fore.GREEN + "Copy the following text to Changelog.rst:")
        print(Fore.YELLOW + header)
        print(Fore.CYAN + output_text)
    else:
        dest.write(final_output + "\n")
        dest.close()


if __name__ == "__main__":
    main()



# Create your tasks here

from demoapp.models import Widget

from celery import shared_task


@shared_task
def add(x, y):
    return x + y


@shared_task
def mul(x, y):
    return x * y


@shared_task
def xsum(numbers):
    return sum(numbers)


@shared_task
def count_widgets():
    return Widget.objects.count()


@shared_task
def rename_widget(widget_id, name):
    w = Widget.objects.get(id=widget_id)
    w.name = name
    w.save()


@shared_task(
    bind=True,
    autoretry_for=(Exception,),
    retry_kwargs={"max_retries": 2, "countdown": 10 * 60},  # retry up to 2 times with 10 minutes between retries
)
def error_task(self):
    raise Exception("Test error")


@shared_task(
    bind=True,
    autoretry_for=(Exception,),
    retry_backoff=5,  # Factor in seconds (first retry: 5s, second: 10s, third: 20s, etc.)
    retry_jitter=False,  # Set False to disable randomization (use exact values: 5s, 10s, 20s)
    retry_kwargs={"max_retries": 3},
)
def error_backoff_test(self):
    raise Exception("Test error")



from uuid import uuid4

from celery.canvas import Signature, StampingVisitor
from celery.utils.log import get_task_logger

logger = get_task_logger(__name__)


class MyStampingVisitor(StampingVisitor):
    def on_signature(self, sig: Signature, **headers) -> dict:
        logger.critical(f"Visitor: Sig '{sig}' is stamped with: mystamp")
        return {"mystamp": "I am a stamp!"}


class MonitoringIdStampingVisitor(StampingVisitor):
    def on_signature(self, sig: Signature, **headers) -> dict:
        mtask_id = str(uuid4())
        logger.critical(f"Visitor: Sig '{sig}' is stamped with: {mtask_id}")
        return {"mtask_id": mtask_id}


class FullVisitor(StampingVisitor):
    def on_signature(self, sig: Signature, **headers) -> dict:
        logger.critical(f"Visitor: Sig '{sig}' is stamped with: on_signature")
        return {
            "on_signature": "FullVisitor.on_signature()",
        }

    def on_callback(self, sig, **headers) -> dict:
        logger.critical(f"Visitor: Sig '{sig}' is stamped with: on_callback")
        return {
            "on_callback": "FullVisitor.on_callback()",
        }

    def on_errback(self, sig, **headers) -> dict:
        logger.critical(f"Visitor: Sig '{sig}' is stamped with: on_errback")
        return {
            "on_errback": "FullVisitor.on_errback()",
        }

    def on_chain_start(self, sig: Signature, **headers) -> dict:
        logger.critical(f"Visitor: Sig '{sig}' is stamped with: on_chain_start")
        return {
            "on_chain_start": "FullVisitor.on_chain_start()",
        }

    def on_group_start(self, sig: Signature, **headers) -> dict:
        logger.critical(f"Visitor: Sig '{sig}' is stamped with: on_group_start")
        return {
            "on_group_start": "FullVisitor.on_group_start()",
        }

    def on_chord_header_start(self, sig: Signature, **headers) -> dict:
        logger.critical(f"Visitor: Sig '{sig}' is stamped with: on_chord_header_start")
        s = super().on_chord_header_start(sig, **headers)
        s.update(
            {
                "on_chord_header_start": "FullVisitor.on_chord_header_start()",
            }
        )
        return s

    def on_chord_body(self, sig: Signature, **headers) -> dict:
        logger.critical(f"Visitor: Sig '{sig}' is stamped with: on_chord_body")
        return {
            "on_chord_body": "FullVisitor.on_chord_body()",
        }



import logging
import time
from concurrent.futures import ThreadPoolExecutor, as_completed

import pytest
from kombu import Exchange, Queue

from celery import Celery, chord
from celery.contrib.testing.worker import start_worker
from celery.result import allow_join_result

logger = logging.getLogger(__name__)


@pytest.fixture(scope="function")
def app():
    """
    Celery app configured to:
    - Use quorum queues with topic exchanges
    - Route chord_unlock to a dedicated quorum queue
    """
    app = Celery(
        "test_app",
        broker="pyamqp://guest:guest@rabbit:5672//",
        backend="redis://redis/0",
    )

    app.conf.task_default_exchange_type = "topic"
    app.conf.task_default_exchange = "default_exchange"
    app.conf.task_default_queue = "default_queue"
    app.conf.task_default_routing_key = "default"

    app.conf.task_queues = [
        Queue(
            "header_queue",
            Exchange("header_exchange", type="topic"),
            routing_key="header_rk",
            queue_arguments={"x-queue-type": "quorum"},
        ),
        Queue(
            "chord_callback_queue",
            Exchange("chord_callback_exchange", type="topic"),
            routing_key="chord_callback_queue",
            queue_arguments={"x-queue-type": "quorum"},
        ),
    ]

    app.conf.task_routes = {
        "celery.chord_unlock": {
            "queue": "chord_callback_queue",
            "exchange": "chord_callback_exchange",
            "routing_key": "chord_callback_queue",
            "exchange_type": "topic",
        },
    }

    return app


@pytest.fixture
def add(app):
    @app.task(bind=True, max_retries=3, default_retry_delay=1)
    def add(self, x, y):
        time.sleep(0.05)
        return x + y
    return add


@pytest.fixture
def summarize(app):
    @app.task(bind=True, max_retries=3, default_retry_delay=1)
    def summarize(self, results):
        return sum(results)
    return summarize


def wait_for_chord_unlock(chord_result, timeout=10, interval=0.2):
    """
    Waits for chord_unlock to be enqueued by polling the `parent` of the chord result.
    This confirms that the header group finished and the callback is ready to run.
    """
    start = time.monotonic()
    while time.monotonic() - start < timeout:
        if chord_result.parent and chord_result.parent.ready():
            return True
        time.sleep(interval)
    return False


@pytest.mark.amqp
@pytest.mark.timeout(90)
@pytest.mark.xfail(reason="chord_unlock routed to quorum/topic queue intermittently fails under load")
def test_chord_unlock_stress_routing_to_quorum_queue(app, add, summarize):
    """
    Reproduces Celery Discussion #9742 (intermittently):
    When chord_unlock is routed to a quorum queue via topic exchange, it may not be consumed
    even if declared and bound, leading to stuck results.

    This stress test submits many chords rapidly, each routed explicitly via a topic exchange,
    and waits to see how many complete.
    """
    chord_count = 50
    header_fanout = 3
    failures = []

    pending_results = []

    with allow_join_result():
        # Submit chords BEFORE worker is running
        for i in range(chord_count):
            header = [
                add.s(i, j).set(
                    queue="header_queue",
                    exchange="header_exchange",
                    routing_key="header_rk",
                )
                for j in range(header_fanout)
            ]

            callback = summarize.s().set(
                queue="chord_callback_queue",
                exchange="chord_callback_exchange",
                routing_key="chord_callback_queue",
            )

            result = chord(header)(callback)
            pending_results.append((i, result))

        # Wait for chord_unlock tasks to be dispatched before starting the worker
        for i, result in pending_results:
            if not wait_for_chord_unlock(result):
                logger.warning(f"[!] Chord {i}: unlock was not dispatched within timeout")

        # Start worker that consumes both header and callback queues
        with start_worker(
            app, queues=["header_queue", "chord_callback_queue"], loglevel="info", perform_ping_check=False
        ):
            # Poll all chord results
            with ThreadPoolExecutor(max_workers=10) as executor:
                futures = {
                    executor.submit(result.get, timeout=20): (i, result)
                    for i, result in pending_results
                }

                for future in as_completed(futures):
                    i, result = futures[future]
                    try:
                        res = future.result()
                        logger.info(f"[✓] Chord {i} completed: {res}")
                    except Exception as exc:
                        logger.error(f"[✗] Chord {i} failed or stuck: {exc}")
                        failures.append((i, exc))

    # Assertion: all chords should have completed
    assert not failures, f"{len(failures)} of {chord_count} chords failed or got stuck"



from time import sleep

from tasks import identity_task, mul, wait_for_revoke, xsum
from visitors import MonitoringIdStampingVisitor

from celery.canvas import Signature, chain, chord, group
from celery.result import AsyncResult


def create_canvas(n: int) -> Signature:
    """Creates a canvas to calculate: n * sum(1..n) * 10
    For example, if n = 3, the result is 3 * (1 + 2 + 3) * 10 = 180
    """
    canvas = chain(
        group(identity_task.s(i) for i in range(1, n+1)) | xsum.s(),
        chord(group(mul.s(10) for _ in range(1, n+1)), xsum.s()),
    )

    return canvas


def revoke_by_headers(result: AsyncResult, terminate: bool) -> None:
    """Revokes the last task in the workflow by its stamped header

    Arguments:
        result (AsyncResult): Can be either a frozen or a running result
        terminate (bool): If True, the revoked task will be terminated
    """
    result.revoke_by_stamped_headers({'mystamp': 'I am a stamp!'}, terminate=terminate)


def prepare_workflow() -> Signature:
    """Creates a canvas that waits "n * sum(1..n) * 10" in seconds,
    with n = 3.

    The canvas itself is stamped with a unique monitoring id stamp per task.
    The waiting task is stamped with different consistent stamp, which is used
    to revoke the task by its stamped header.
    """
    canvas = create_canvas(n=3)
    canvas = canvas | wait_for_revoke.s()
    canvas.stamp(MonitoringIdStampingVisitor())
    return canvas


def run_then_revoke():
    """Runs the workflow and lets the waiting task run for a while.
    Then, the waiting task is revoked by its stamped header.

    The expected outcome is that the canvas will be calculated to the end,
    but the waiting task will be revoked and terminated *during its run*.

    See worker logs for more details.
    """
    canvas = prepare_workflow()
    result = canvas.delay()
    print('Wait 5 seconds, then revoke the last task by its stamped header: "mystamp": "I am a stamp!"')
    sleep(5)
    print('Revoking the last task...')
    revoke_by_headers(result, terminate=True)


def revoke_then_run():
    """Revokes the waiting task by its stamped header before it runs.
    Then, run the workflow, which will not run the waiting task that was revoked.

    The expected outcome is that the canvas will be calculated to the end,
    but the waiting task will not run at all.

    See worker logs for more details.
    """
    canvas = prepare_workflow()
    result = canvas.freeze()
    revoke_by_headers(result, terminate=False)
    result = canvas.delay()



import os
import pickle
import sys
import tempfile
import time
from unittest.mock import patch

import pytest

import t.skip
from celery import states, uuid
from celery.backends import filesystem
from celery.backends.filesystem import FilesystemBackend
from celery.exceptions import ImproperlyConfigured


@t.skip.if_win32
class test_FilesystemBackend:

    def setup_method(self):
        self.directory = tempfile.mkdtemp()
        self.url = 'file://' + self.directory
        self.path = self.directory.encode('ascii')

    def test_a_path_is_required(self):
        with pytest.raises(ImproperlyConfigured):
            FilesystemBackend(app=self.app)

    def test_a_path_in_url(self):
        tb = FilesystemBackend(app=self.app, url=self.url)
        assert tb.path == self.path

    @pytest.mark.parametrize("url,expected_error_message", [
        ('file:///non-existing', filesystem.E_PATH_INVALID),
        ('url://non-conforming', filesystem.E_PATH_NON_CONFORMING_SCHEME),
        (None, filesystem.E_NO_PATH_SET)
    ])
    def test_raises_meaningful_errors_for_invalid_urls(
        self,
        url,
        expected_error_message
    ):
        with pytest.raises(
            ImproperlyConfigured,
            match=expected_error_message
        ):
            FilesystemBackend(app=self.app, url=url)

    def test_localhost_is_removed_from_url(self):
        url = 'file://localhost' + self.directory
        tb = FilesystemBackend(app=self.app, url=url)
        assert tb.path == self.path

    def test_missing_task_is_PENDING(self):
        tb = FilesystemBackend(app=self.app, url=self.url)
        assert tb.get_state('xxx-does-not-exist') == states.PENDING

    def test_mark_as_done_writes_file(self):
        tb = FilesystemBackend(app=self.app, url=self.url)
        tb.mark_as_done(uuid(), 42)
        assert len(os.listdir(self.directory)) == 1

    def test_done_task_is_SUCCESS(self):
        tb = FilesystemBackend(app=self.app, url=self.url)
        tid = uuid()
        tb.mark_as_done(tid, 42)
        assert tb.get_state(tid) == states.SUCCESS

    def test_correct_result(self):
        data = {'foo': 'bar'}

        tb = FilesystemBackend(app=self.app, url=self.url)
        tid = uuid()
        tb.mark_as_done(tid, data)
        assert tb.get_result(tid) == data

    def test_get_many(self):
        data = {uuid(): 'foo', uuid(): 'bar', uuid(): 'baz'}

        tb = FilesystemBackend(app=self.app, url=self.url)
        for key, value in data.items():
            tb.mark_as_done(key, value)

        for key, result in tb.get_many(data.keys()):
            assert result['result'] == data[key]

    def test_forget_deletes_file(self):
        tb = FilesystemBackend(app=self.app, url=self.url)
        tid = uuid()
        tb.mark_as_done(tid, 42)
        tb.forget(tid)
        assert len(os.listdir(self.directory)) == 0

    @pytest.mark.usefixtures('depends_on_current_app')
    def test_pickleable(self):
        tb = FilesystemBackend(app=self.app, url=self.url, serializer='pickle')
        assert pickle.loads(pickle.dumps(tb))

    @pytest.mark.skipif(sys.platform == 'win32', reason='Test can fail on '
                        'Windows/FAT due to low granularity of st_mtime')
    def test_cleanup(self):
        tb = FilesystemBackend(app=self.app, url=self.url)
        yesterday_task_ids = [uuid() for i in range(10)]
        today_task_ids = [uuid() for i in range(10)]
        for tid in yesterday_task_ids:
            tb.mark_as_done(tid, 42)
        day_length = 0.2
        time.sleep(day_length)  # let FS mark some difference in mtimes
        for tid in today_task_ids:
            tb.mark_as_done(tid, 42)
        with patch.object(tb, 'expires', 0):
            tb.cleanup()
        # test that zero expiration time prevents any cleanup
        filenames = set(os.listdir(tb.path))
        assert all(
            tb.get_key_for_task(tid) in filenames
            for tid in yesterday_task_ids + today_task_ids
        )
        # test that non-zero expiration time enables cleanup by file mtime
        with patch.object(tb, 'expires', day_length):
            tb.cleanup()
        filenames = set(os.listdir(tb.path))
        assert not any(
            tb.get_key_for_task(tid) in filenames
            for tid in yesterday_task_ids
        )
        assert all(
            tb.get_key_for_task(tid) in filenames
            for tid in today_task_ids
        )
